#!/usr/bin/env python
# -*- coding: utf-8 -*-

from gensim.utils import SaveLoad
from sekg.so import SOTagItem


class SOTagItemCollection(SaveLoad):
    def __init__(self):
        self.name2sotag = {}
        self.synonyms2tagname = {}

    def update_synonyms2tagname(self, so_tag_item: SOTagItem):
        if so_tag_item.synonyms:
            for synonym in so_tag_item.synonyms:
                self.synonyms2tagname[synonym] = so_tag_item.tag_name

    def add_so_tag_item(self, so_tag_item: SOTagItem):
        if so_tag_item is None:
            return False
        self.name2sotag[so_tag_item.tag_name] = so_tag_item
        self.update_synonyms2tagname(so_tag_item)
        return True

    def get_so_tag_item(self, tag_name):
        if tag_name in self.name2sotag.keys():
            return self.name2sotag.get(tag_name, None)
        else:
            if tag_name in self.synonyms2tagname:
                return self.name2sotag.get(self.synonyms2tagname[tag_name], None)
        return None

    def sub_collection(self, tag_names):
        subcollection = SOTagItemCollection()
        valid_tag_name = self.get_tag_names()
        for tag_name in tag_names:
            if tag_name in valid_tag_name:
                subcollection.add_so_tag_item(self.get_so_tag_item(tag_name))
        return subcollection

    def size(self):
        return len(set(self.name2sotag.keys()) | set(self.name2sotag.keys()))

    def get_tag_names(self):
        return set(self.name2sotag.keys())

    def get_all_tag_name(self):
        return set(self.name2sotag.keys()) | set(self.name2sotag.keys())



from sekg.text.spacy_pipeline.pipeline import PipeLineFactory
from sekg.util.dependency_tree_util import DependencyTreeUtil


class CompleteSubject:
    def __init__(self, nlp=None):
        if nlp is None:
            self.nlp = PipeLineFactory.full_pipeline()
        else:
            self.nlp = nlp

    def clean_pure_code_sentence(self, text: str):
        text = text.strip()
        if text.find(" ") < 0:
            return ""
        if text.find(".java") >= 0:
            return ""
        if text.find(">") >= 0 and text.find("<") >= 0 and text.find(";") >= 0:
            return ""
        if text.find("=") >= 0:
            return ""
        if text.find("(") >= 0 and text.find(";") >= 0:
            return ""

        return text

    def get_sentence_list_with_doc(self, doc):
        l = []
        for sent in doc.sents:
            clean = self.clean_pure_code_sentence(sent.text)
            if clean == "":
                continue
            l.append(sent)
        return l

    def complete_subject_by_name_for_doc(self, full_text, candidate_subject_name, spacy_doc=None,
                                         candidate_verb_set=None):
        """
        candidate_verb_set是从类下面所有的method名称或者method自己的名字里面抽取出的所有动词
        :param full_text:
        :param candidate_subject_name:
        :param spacy_doc:
        :param candidate_verb_set:
        :return:
        """
        if spacy_doc is not None:
            test_doc = spacy_doc
        else:
            test_doc = self.nlp(full_text)
        result = []
        for index, sent in enumerate(test_doc.sents):
            if index == 0:
                ## for long text, only try to complete the first sentence.
                result.append(
                    self.complete_subject_for_sentence(sent.text, candidate_subject_name, sent, candidate_verb_set))
            else:
                result.append(sent.text)

        return " ".join(result)

    def how_many_upper(self, input_text):
        """
        有多少个大写字符
        :param input_text:
        :return:
        """
        n = 0
        for l, r in zip(input_text.lower(), input_text):
            if l != r:
                n += 1
        return n

    def is_declarative_sentence(self, sentence):
        """
        check whether a sentence is
        :return:
        """
        # todo: complete this method to consider more situation
        if sentence[-1] == "?" or sentence[-1] == "!":
            return False
        return True

    def complete_subject_for_sentence(self, sentence, candidate_subject_name, sent_doc=None, candidate_verb_set=None):
        """

        :param candidate_verb_set:
        :param sent_doc:
        :param sentence:
        :param candidate_subject_name:
        :return:补全后的句子
        """
        if not self.is_declarative_sentence(sentence):
            return sentence
        if candidate_verb_set is None:
            candidate_verb_set = set()
        candidate_verb_set.add("return")
        candidate_verb_set.add("create")
        candidate_verb_set.add("construct")
        candidate_verb_set.add("get")
        candidate_verb_set.add("set")
        subject = DependencyTreeUtil.get_subject_exclude_verb(sent_doc, candidate_verb_set)
        # for subject completor, it only check the sentence, it assume the sentence parsing is correct.
        if subject is not None:
            return sentence

        predicate = DependencyTreeUtil.get_main_predicate_with_candidate_verbs(sent_doc, candidate_verb_set)
        if sent_doc[0].pos_ == "ADV" and sent_doc[0].tag_ == "EX":
            return sentence
        # is_vbn = DependencyTreeUtil.is_vbn(sent_doc)
        if predicate is None:
            append_str = candidate_subject_name + " is "
        else:
            append_str = candidate_subject_name + " "

        f = str(sentence[0]).lower()
        first_token = sent_doc[0]
        if self.how_many_upper(first_token.text) >= 1:
            sentence = append_str + f + sentence[1:]
        else:
            sentence = append_str + sentence

        return sentence



import traceback

from sqlalchemy import Column, Integer, String, Text, MetaData, ForeignKey, Index, func, or_
from sqlalchemy.dialects.mysql import LONGTEXT
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from sekg.constant.code import CodeEntityCategory, CodeEntityRelationCategory

Base = declarative_base()
metadata = MetaData()


# api_relation_table = Table(
#     'java_api_relation', metadata,
#     Column("id", Integer, primary_key=True, autoincrement=True),
#     Column("start_api_id", Integer, ForeignKey('java_all_api_entity.id'), nullable=False, index=True),
#     Column("end_api_id", Integer, ForeignKey('java_all_api_entity.id'), nullable=False, index=True),
#     Column("relation_type", Integer, index=True),
# )
# api_entity_table = Table(
#     'java_all_api_entity', metadata,
#     Column("id", Integer, primary_key=True, autoincrement=True),
#     Column("api_type", Integer, index=True),
#     Column("qualified_name", String(1024), index=True),
#     Column("full_declaration", String(1024), nullable=True, index=True),
#     Column("short_description", Text(), nullable=True),
#     Column("added_in_version", String(128), nullable=True),
# )


class APIRelation(Base):
    __tablename__ = 'java_api_relation'
    id = Column(Integer, primary_key=True, autoincrement=True)
    start_api_id = Column(Integer, ForeignKey('java_all_api_entity.id'), nullable=False, index=True)
    end_api_id = Column(Integer, ForeignKey('java_all_api_entity.id'), nullable=False, index=True)
    relation_type = Column(Integer, index=True)

    __table_args__ = (Index('unique_index', start_api_id, end_api_id, relation_type),
                      Index('all_relation_index', start_api_id, end_api_id),
                      {
                          "mysql_charset": "utf8",
                      })

    def __init__(self, start_api_id, end_api_id, relation_type):
        self.start_api_id = start_api_id
        self.end_api_id = end_api_id
        self.relation_type = relation_type

    def exist_in_remote(self, session):
        try:
            if session.query(APIRelation.id).filter_by(start_api_id=self.start_api_id,
                                                       end_api_id=self.end_api_id,
                                                       relation_type=self.relation_type).first():
                return True
            else:
                return False
        except Exception:
            traceback.print_exc()
            return False

    def get_remote_object(self, session):
        if self.id:
            return self
        else:
            try:
                return session.query(APIRelation).filter_by(start_api_id=self.start_api_id,
                                                            end_api_id=self.end_api_id,
                                                            relation_type=self.relation_type).first()
            except Exception:
                # traceback.print_exc()
                return None

    def find_or_create(self, session, autocommit=True):
        remote_instance = self.get_remote_object(session)
        if not remote_instance:
            session.add(self)
            if autocommit:
                session.commit()
            return self
        else:
            return remote_instance

    def __repr__(self):
        return '<APIRelation: %r-%r: type=%r >' % (self.start_api_id, self.end_api_id, self.relation_type)

    @staticmethod
    def get_api_relation_by_start_and_end_api_id(session, start_api_id, end_api_id):
        try:
            api_relation = session.query(APIRelation).filter_by(start_api_id=start_api_id,
                                                                end_api_id=end_api_id).first()
            return api_relation
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_type_string(relation_type):
        if relation_type in CodeEntityRelationCategory.category_code_to_str_map:
            return CodeEntityRelationCategory.category_code_to_str_map[relation_type]
        else:
            return ""

    @staticmethod
    def get_relation_by_relation_type(session, relation_type):
        try:
            return session.query(APIRelation).filter_by(relation_type=relation_type).all()
        except Exception:
            traceback.print_exc()
            return None


class APIEntity(Base):
    __tablename__ = 'java_all_api_entity'
    id = Column(Integer, primary_key=True, autoincrement=True)
    api_type = Column(Integer, index=True)
    qualified_name = Column(String(1024), index=True)
    full_declaration = Column(String(1024), nullable=True, index=True)
    short_description = Column(Text(), nullable=True)
    added_in_version = Column(String(128), nullable=True)

    out_relation = relationship('APIRelation', foreign_keys=[APIRelation.start_api_id],
                                backref='start_api')
    in_relation = relationship('APIRelation', foreign_keys=[APIRelation.end_api_id],
                               backref='end_api')

    __table_args__ = {
        "mysql_charset": "utf8"
    }

    def __init__(self, qualified_name, api_type, full_declaration=None, short_description=None, added_in_version=None):
        self.api_type = api_type
        self.qualified_name = qualified_name
        self.full_declaration = full_declaration
        self.short_description = short_description
        self.added_in_version = added_in_version

    def find_or_create(self, session, autocommit=True):
        if self.api_type == CodeEntityCategory.CATEGORY_PARAMETER or self.api_type == CodeEntityCategory.CATEGORY_RETURN_VALUE or self.api_type == CodeEntityCategory.CATEGORY_EXCEPTION_CONDITION:
            remote_instance = self.get_remote_parameter_object(session)
        else:
            remote_instance = self.get_remote_object(session)
        if not remote_instance:
            session.add(self)
            if autocommit:
                session.commit()
            return self
        else:
            return remote_instance

    def get_remote_object(self, session):
        if self.id:
            return self
        else:
            try:
                return session.query(APIEntity).filter(
                    APIEntity.qualified_name == func.binary(self.qualified_name)).first()
            except Exception:
                traceback.print_exc()
                return None

    def get_remote_parameter_object(self, session):
        if self.id:
            return self
        else:
            try:
                return session.query(APIEntity).filter_by(qualified_name=self.qualified_name,
                                                          full_declaration=self.full_declaration,
                                                          short_description=self.short_description).first()
            except Exception:
                traceback.print_exc()
                return None

    @staticmethod
    def exist(session, qualified_name):
        try:
            if session.query(APIEntity.id).filter(APIEntity.qualified_name == func.binary(qualified_name)).first():
                return True
            else:
                return False
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def find_by_id(session, api_entity_id):
        try:
            return session.query(APIEntity).filter(APIEntity.id == api_entity_id).first()
        except Exception:
            return None

    @staticmethod
    def find_by_qualifier(session, qualified_name):
        try:
            return session.query(APIEntity).filter(APIEntity.qualified_name == func.binary(qualified_name)).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def find_by_full_declaration_and_description(session, full_declaration, description):
        try:
            return session.query(APIEntity).filter_by(full_declaration=func.binary(full_declaration),
                                                      short_description=description).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_api_type_string(type):
        if type in CodeEntityCategory.category_code_to_str_list_map:
            return CodeEntityCategory.category_code_to_str_list_map[type]
        return []

    @staticmethod
    def get_simple_type_string(type):
        if type in CodeEntityCategory.category_code_to_str_map:
            return CodeEntityCategory.category_code_to_str_map[type]
        return ""

    def __repr__(self):
        return '<APIEntity: id=%r name=%r>' % (self.id, self.qualified_name)

    def __eq__(self, other):
        if isinstance(other, APIEntity):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        return hash(self.id)

    @staticmethod
    def get_api_id_list(session):
        try:
            return session.query(APIEntity.id).all()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_api_id_and_qualified_name_list(session):
        try:
            return session.query(APIEntity.id, APIEntity.qualified_name).all()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_all_API_entity(session):
        try:
            return session.query(APIEntity).all()
        except Exception:
            traceback.print_exc()
            return []

    @staticmethod
    def get_all_value_instance_api(session):
        try:
            return session.query(APIEntity).filter(
                or_(APIEntity.api_type == CodeEntityCategory.CATEGORY_EXCEPTION_CONDITION,
                    APIEntity.api_type == CodeEntityCategory.CATEGORY_PARAMETER,
                    APIEntity.api_type == CodeEntityCategory.CATEGORY_RETURN_VALUE)).all()
        except Exception:
            traceback.print_exc()
            return []

    @staticmethod
    def get_qualified_by_id(session, id):
        try:
            return session.query(APIEntity.qualified_name).filter_by(id=id).first()
        except Exception:
            traceback.print_exc()
            return None


class APIHTMLText(Base):
    __tablename__ = 'java_api_html_text'
    id = Column(Integer, primary_key=True, autoincrement=True)
    api_id = Column(Integer, ForeignKey('java_all_api_entity.id'), nullable=False)
    html = Column(LONGTEXT(), nullable=False)
    clean_text = Column(LONGTEXT(), nullable=True)  # text with no html tags
    reserve_part_tag_text = Column(LONGTEXT(), nullable=True)  # text with only code tags text
    html_type = Column(Integer, nullable=True)

    __table_args__ = (Index('api_id_text_type_index', api_id, html_type), {
        "mysql_charset": "utf8",
    })

    HTML_TYPE_UNKNOWN = 0
    HTML_TYPE_API_DECLARATION = 1
    HTML_TYPE_API_SHORT_DESCRIPTION = 2
    HTML_TYPE_API_DETAIL_DESCRIPTION = 3
    HTML_TYPE_METHOD_RETURN_VALUE_DESCRIPTION = 4

    def __init__(self, api_id, html, html_type=HTML_TYPE_UNKNOWN):
        self.api_id = api_id
        self.html = html
        self.html_type = html_type

    def create(self, session, autocommit=True):
        session.add(self)
        if autocommit:
            session.commit()
        return self

    @staticmethod
    def get_by_id(session, id):
        try:
            return session.query(APIHTMLText).filter_by(id=id).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_text_by_api_id_and_type(session, api_id, html_type):
        try:
            return session.query(APIHTMLText.clean_text).filter_by(api_id=api_id, html_type=html_type).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_remote_object(session, api_id, html_type):
        try:
            return session.query(APIHTMLText).filter_by(api_id=api_id, html_type=html_type).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_html_text_id(session, api_id, html_type):
        try:
            return session.query(APIHTMLText.id).filter_by(api_id=api_id, html_type=html_type).first()
        except Exception:
            traceback.print_exc()
            return None

    @staticmethod
    def get_all_html_text_by_html_type(session, html_type):
        try:
            return session.query(APIHTMLText).filter_by(html_type=html_type).all()
        except Exception:
            traceback.print_exc()
            return None

    def find_or_create(self, session, autocommit=True):
        remote_instance = self.get_remote_object(session, api_id=self.api_id, html_type=self.html_type)
        if not remote_instance:
            session.add(self)
            if autocommit:
                session.commit()
            return self
        else:
            return remote_instance



from unittest import TestCase

from sekg.so import SOTagItem

from sekg.so.stackoverflow_tag_searcher import SOTagSearcher


class TestSoTagSearcher(TestCase):

    def test_so_tag_searcher(self):
        searcher = SOTagSearcher()
        print(len(searcher.synonyms_data))
        so_tag_item: SOTagItem = searcher.get_tag_item_for_one_tag("javascript")
        print(so_tag_item.tag_name)
        print(so_tag_item.synonyms)
        print(so_tag_item.short_description)
        searcher.run()
        searcher.save("test.bin")




"""
this class is for analysis the sentence based on dependency tree.

"""
import spacy
from spacy.tokens.doc import Doc

from sekg.util.spacy_fixer import SoftwareTextPOSFixer
from sekg.util.vocabulary_conversion.vocabulary_conversion import VocabularyConversion

from sekg.text.spacy_pipeline.sentenceHandler import SentenceHandler


class DependencyTreeUtil:
    nlp = spacy.load('en', disable=["ner"])
    nlp.add_pipe(SentenceHandler.hyphen_handler, name='sh', before='tagger')
    # 将核心动词与紧挨着的head是这个动词的介词合并
    # nlp.add_pipe(SentenceHandler.merge_verb_prep, name='merge_verb_prep', after='parser')
    nlp.add_pipe(SoftwareTextPOSFixer.fixer_for_pos, name="pos_fixer", after="tagger")

    @staticmethod
    def get_attr_for_be_predicate(doc, predicate_token=None):
        """
        主要是获取主系表中的表语，形容词或者名词，可以用来抽
        the input is the sentence Spacy Doc
        :param doc:
        :return: Span noun chunk for the subject, None for not found.
        """
        if predicate_token == None:
            predicate_token = DependencyTreeUtil.get_main_predicate(doc)
        if predicate_token == None:
            return None
        for possible_attr in doc:
            if possible_attr.dep_ == "attr" and possible_attr.head == predicate_token:
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_attr)
            if possible_attr.dep_ == "acomp" and possible_attr.head == predicate_token:
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_attr)

    @staticmethod
    def split_span_into_parallel(doc, span):
        span_list = []
        if span is None or span.text == "":
            return span_list
        try:
            span_root = span.root
            other_parallel_root_tokens = []
            for token in doc:
                if token.dep_ == "appos" and token.head == span_root:
                    other_parallel_root_tokens.append(token)
                if token.dep_ == "cc" and token.head == span_root:
                    other_parallel_root_tokens.append(token)
            # if len(other_parallel_root_tokens) == 0:
            #     span_list.append(span)
            #     return span_list
            for t in [span_root] + other_parallel_root_tokens:
                left_sub_tree_span = DependencyTreeUtil.get_left_subtree_and_itself_span(doc=doc, token=t)
                span_list.append(left_sub_tree_span)
        except Exception as e:
            print(e)
        return span_list

    @staticmethod
    def get_left_subtree_and_itself_span(doc, token):
        lefts = [t for t in token.lefts]
        index = 0
        pre_index = 0
        for i, or_token in enumerate(doc):
            if or_token.text == token.text:
                index = i
            if len(lefts) > 0:
                if or_token.text == lefts[0].text:
                    pre_index = i
        if len(lefts) == 0:
            return doc[index:index + 1]
        return doc[pre_index:(index + 1)]

    @staticmethod
    def get_main_predicate(doc):
        root = [token for token in doc if token.dep_ == "ROOT"][0]
        if root.pos_ == "VERB" or root.pos_ == 'ADJ' or root.pos_ == 'AUX':
            return root
        else:
            return None

    @staticmethod
    def get_conj_predicate(doc):
        token_list = []
        for token in doc:
            if token.pos_ == "VERB" and token.head.tag_ == "cc" and token.head.head.dep_ == "ROOT":
                token_list.append(token)
        # token = [token for token in doc if token.pos_ == "VERB" and token.head == "cc" and token.head.head.dep_ == "ROOT"]
        return token_list

    @staticmethod
    def get_main_predicate_with_candidate_verbs(doc, verb_set):
        root = [token for token in doc if token.dep_ == "ROOT"][0]
        for i, token in enumerate(doc):
            if i == 0 and (token.lemma_ in verb_set or VocabularyConversion.couldBeVerb_probability(token.text)):
                return token
            if token.dep_ == "ROOT":
                root = token
                break

        if root.pos_ == "VERB" or (root.lemma_ is not None and root.lemma_ in verb_set):
            return root
        else:
            return None

    @staticmethod
    def is_vbn(doc):
        root = [token for token in doc if token.dep_ == "ROOT"][0]
        if root.pos_ == "VERB":
            if root.tag_ == 'VBN':
                return True
            return False
        else:
            return False

    @staticmethod
    def get_subject(doc):
        """
        the input is the sentence Spacy Doc
        :param doc:
        :return: Span noun chunk for the subject, None for not found.
        """
        for possible_subject in doc:
            if possible_subject.dep_ == "nsubj" and possible_subject.head.dep_ == "ROOT":
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_subject)

            if possible_subject.dep_ == "nsubjpass" and possible_subject.head.dep_ == "ROOT":
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_subject)

        # todo: need to improve for finding more situation. including some error about software text.

        predicate = DependencyTreeUtil.get_main_predicate(doc)
        ## the sentence is lack of the subject and start with verb. eg. "go to school".
        ## the problem of "sort a given number list" must be fixed.
        if predicate is not None and predicate.n_lefts == 0:
            return None

        root = [token for token in doc if token.dep_ == "ROOT"][0]

        if root.n_lefts == 0:
            return None
        else:
            return

    @staticmethod
    def get_subject_exclude_verb(doc, candidate_verb_set):
        """
        the input is the sentence Spacy Doc
        :return: Span noun chunk for the subject, None for not found.
        """
        for possible_subject in doc:
            if possible_subject.dep_ == "nsubj" and possible_subject.head.dep_ == "ROOT" and possible_subject.lemma_ not in candidate_verb_set:
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_subject)

            if possible_subject.dep_ == "nsubjpass" and possible_subject.head.dep_ == "ROOT" and possible_subject.lemma_ not in candidate_verb_set:
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=possible_subject)

        # todo: need to improve for finding more situation. including some error about software text.

        predicate = DependencyTreeUtil.get_main_predicate(doc)
        ## the sentence is lack of the subject and start with verb. eg. "go to school".
        ## the problem of "sort a given number list" must be fixed.
        if predicate is not None and predicate.n_lefts == 0:
            return None

        root = [token for token in doc if token.dep_ == "ROOT"][0]

        if root.n_lefts == 0:
            return None
        else:
            return None

    @staticmethod
    def get_subject_text(doc):

        subject = DependencyTreeUtil.get_subject(doc)
        if subject == None:
            return None
        return subject.text

    @staticmethod
    def get_subtree_span_from_one_token_obj(doc, token):
        if token == None:
            return None
        # for token_ in doc:
        #     if token_.lemma_ == token.lemma_:
        #         token = token_
        token_list = []
        DependencyTreeUtil.get_all_childern_token_under_one_token(doc, token, token_list)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[token.left_edge.i: token.right_edge.i + 1]
        return span[0: len(span) + 1]

    @staticmethod
    def get_subtree_span_from_one_token_obj_by_index(doc, token, i):
        if token == None:
            return None
        if token.text != doc[i].text:
            return None
        token_list = []
        DependencyTreeUtil.get_all_childern_token_under_one_token(doc, token, token_list)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[token.left_edge.i: token.right_edge.i + 1]
        return span[0: len(span) + 1]

    @staticmethod
    def get_subtree_span_from_token_by_index(doc, token):
        if token == None:
            return None
        span = doc[token.left_edge.i: token.right_edge.i + 1]
        return span

    @staticmethod
    def get_all_childern_token_under_one_token(doc, token, token_list):
        if token.lefts:
            for left in token.lefts:
                DependencyTreeUtil.get_all_childern_token_under_one_token(doc, left, token_list)
        token_list.append(token)
        if token.rights:
            for right in token.rights:
                DependencyTreeUtil.get_all_childern_token_under_one_token(doc, right, token_list)

    @staticmethod
    def get_all_childern_token_under_one_token_filter_one(doc, token, token_list, filter):
        if token.text == filter.text and token.tag_[:2] == filter.tag_[
                                                           :2] and token.left_edge.text == filter.left_edge.text and token.right_edge.text == filter.right_edge.text:
            # print('subject.root')
            # print(token.text)
            pass
        else:
            if token.lefts:
                for left in token.lefts:
                    DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one(doc, left, token_list, filter)
            token_list.append(token)
            if token.rights:
                for right in token.rights:
                    DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one(doc, right, token_list, filter)

    @staticmethod
    def get_all_childern_token_under_one_token_filter_one_for_split_sentence(doc, token, token_list, filter):
        if token.text == filter.text and token.tag_ == filter.tag_ and token.dep_ == filter.dep_ and token.head.text == filter.head.text:
            # print('subject.root')
            # print(token.text)
            pass
        else:
            if token.lefts:
                for left in token.lefts:
                    DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one_for_split_sentence(doc, left,
                                                                                                            token_list,
                                                                                                            filter)
            token_list.append(token)
            if token.rights:
                for right in token.rights:
                    DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one_for_split_sentence(doc, right,
                                                                                                            token_list,
                                                                                                            filter)

    @staticmethod
    def get_subtree_span_from_one_token_obj_for_vbn_and_root(doc, subject, token):
        if token is None:
            return None
        token_list = list()
        # for token_ in doc:
        #     if token_.lemma_ == token.lemma_:
        #         token = token_
        # print('subject.root.text')
        # print(subject.root)
        DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one(doc, token, token_list, subject.root)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[subject.root.right_edge.i+1: token.right_edge.i + 1]
        return span[0: len(span) + 1]

    @staticmethod
    def get_subtree_span_token_under_one_token_filter_one_for_split_sentence(doc, filter, token):
        if token == None:
            return None
        token_list = list()
        # for token_ in doc:
        #     if token_.lemma_ == token.lemma_:
        #         token = token_
        # print('subject.root.text')
        # print(subject.root)
        DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one_for_split_sentence(doc, token, token_list,
                                                                                                filter)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[subject.root.right_edge.i+1: token.right_edge.i + 1]
        return span[0: len(span) + 1]

    @staticmethod
    def get_token_children(doc, token):
        token_list = list()
        for token_ in doc:
            if token_.head == token:
                token_list.append(token_)
        if not token_list:
            return token
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        return span[0: len(span) + 1]

    @staticmethod
    def filter_front_prep_subtree(doc, predicate):
        return doc[predicate.i:]

    @staticmethod
    def swap_condition_to_end(doc, predicate):
        for token in doc:
            if token.dep_.startswith('aux') and token.head.text == predicate.text and token.head.dep_ == 'ROOT':
                swap_span = doc[0: token.i].text
                preidcate_span = doc[token.i:].text
                return preidcate_span.strip('.') + ' $$ ' + swap_span
            if token.head.text == predicate.head.text and token.text == predicate.text and predicate.dep_ == token.dep_:
                swap_span = doc[0: token.i].text
                # if token.i > 0:
                #     if doc[token.i - 1].dep_ == 'aux' and doc[token.i - 1].tag_ == 'MD':
                #         if token.i > 1:
                #             swap_span = doc[0: token.i - 1].text
                #             preidcate_span = doc[token.i - 1:].text
                #             return preidcate_span + ' ' + swap_span
                #     else:
                #         swap_span = doc[0: token.i].text

                preidcate_span = doc[token.i:].text
                return preidcate_span.strip('.') + ' $$ ' + swap_span
        return ""

    @staticmethod
    def get_subtree_span_from_one_token_filter_another_token_sub_tree(doc, filter, token):
        if token == None:
            return None
        token_list = list()

        # print('subject.root.text')
        # print(subject.root)
        # for token_ in doc:
        #     if token_.lemma_ == token.lemma_:
        #         token = token_
        DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one(doc, token, token_list, filter)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[subject.root.right_edge.i+1: token.right_edge.i + 1]
        return span[0: len(span) + 1]

    @staticmethod
    def get_subtree_span_from_one_token_filter_another_token_sub_tree_for_split_sentence(doc, filter, token):
        if token == None:
            return None
        token_list = list()

        # print('subject.root.text')
        # print(subject.root)
        # for token_ in doc:
        #     if token_.lemma_ == token.lemma_:
        #         token = token_
        DependencyTreeUtil.get_all_childern_token_under_one_token_filter_one_for_split_sentence(doc, token, token_list,
                                                                                                filter)
        span_text = ' '.join([token.text for token in token_list])
        span = DependencyTreeUtil.nlp(span_text)
        span = DependencyTreeUtil.merge_np_chunks(span)
        span = DependencyTreeUtil.merge_np_of_np(span)
        # span = doc[subject.root.right_edge.i+1: token.right_edge.i + 1]
        if span:
            return span[0: len(span) + 1]
        else:
            return None

    @staticmethod
    def get_subtree_span_from_functional_verb(doc, token):
        if token is None:
            return None

        span = doc[token.i + 1: token.right_edge.i + 1]
        return span

    @staticmethod
    def get_subtree_span_from_one_token_index(doc, token_index):
        token = DependencyTreeUtil.get_token_by_index(doc, token_index=token_index)
        return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token)

    @staticmethod
    def get_token_by_index(doc, token_index):
        if token_index < 0 or token_index > len(doc):
            return None
        token = doc[token_index]
        return token

    @staticmethod
    def get_action_text_for_token(doc, token):
        """
        抽取抽出来的动词作用对象，比如parse A to B， 抽出to B
        :param doc:
        :param token:
        :return:
        """
        result = []
        for candiate_candition_root_token in doc:
            try:
                if candiate_candition_root_token.dep_ == "prep" and candiate_candition_root_token.head == token:
                    span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                                  token=candiate_candition_root_token)
                    result.append(span.text)
            except Exception as e:
                print(e)
        return result

    @staticmethod
    def get_conditions_text_for_token(doc, token):
        """
        get the conditions, eg. StringBuffer could be modified during single thread at one time.
        when given the "modified" as root, the "during single thread at one time." should be return as condition.

        :param doc:
        :param token:
        :return:
        """
        result = []
        for candiate_candition_root_token in doc:
            ## 使用 介词(prep)+短语作为条件
            if candiate_candition_root_token.tag_ == "RBR":
                continue
            if candiate_candition_root_token.dep_ == "prep" and candiate_candition_root_token.head.text == token.text:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=candiate_candition_root_token)
                result.append(span.text)

            ## 使用when+从句作为条件
            if candiate_candition_root_token.dep_ == "advcl" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=candiate_candition_root_token)
                result.append(span.text)

            ## 使用when+从句作为条件
            if candiate_candition_root_token.dep_ == "advmod" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                if len(span) > 1:
                    result.append(span.text)

            ## 使用that+从句作为条件
            if candiate_candition_root_token.dep_ == "ccomp" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)

        if len(result) == 0:
            return ""

        return " ".join(result)

    @staticmethod
    def get_conditions_for_based_on_some_center_node_index_as_text(doc, token_index):
        """
        get the conditions, eg. StringBuffer could be modified during single thread at one time.
        when given the "modified" as root, the "during single thread at one time." should be return as condition.

        :param doc:
        :param token:
        :return:
        """
        token = DependencyTreeUtil.get_token_by_index(doc, token_index)

        return DependencyTreeUtil.get_conditions_text_for_token(doc, token)

    @staticmethod
    def split_large_noun_phase_span_to_adj_and_np(span):
        """
        将一个名词的span切割成多个修饰它的定语与名词，主要服务于特征和category抽取。
        :param span:
        :return:
        """
        root_token = [t for t in span if t.dep_ == "ROOT"][0]
        adj_list = DependencyTreeUtil.get_adj_modifier_for_noun(span, root_token)
        text = span.text
        for adj in adj_list:
            text = text.replace(adj, " ")
        text = text.replace(",", " ")
        # print("replace=", text)

        # todo: fix this,这个是为了取出句子中的主要名词，应该有其他方法。
        new_words = []
        words = text.split(" ")
        for word in words:
            if word.lower() in ["a", "an", "the", ","]:
                continue
            if word == " " or word == "":
                continue
            new_words.append(word)
        text = " ".join(new_words)
        return text, set(adj_list)

    @staticmethod
    def get_adj_modifier_for_noun(doc, token):
        """
        获取修饰名词的形容词部分
        :param doc:
        :param token:
        :return:
        """

        result = []
        for candiate_candition_root_token in doc:
            if candiate_candition_root_token.dep_ == "acl" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token=candiate_candition_root_token)
                result.append(span.text)

            if candiate_candition_root_token.dep_ == "amod" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)

            ## 使用that+从句作为条件
            if candiate_candition_root_token.dep_ == "ccomp" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)
                ## 使用that+从句作为条件
            if candiate_candition_root_token.dep_ == "relcl" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)

            if candiate_candition_root_token.dep_ == "advmod" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)

            if candiate_candition_root_token.dep_ == "advcl" and candiate_candition_root_token.head == token:
                span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc,
                                                                              token=candiate_candition_root_token)
                result.append(span.text)

        if len(result) == 0:
            return []

        return result

    @staticmethod
    def merge_np_chunks(doc: Doc):
        """
        merge the np chunks in to one token for the the better sentence analysis and extractor.
        :param doc:
        :return:
        """
        with doc.retokenize() as retokenizer:
            for chunk in doc.noun_chunks:
                try:
                    attrs = {"LEMMA": chunk[:-1].text + " " + chunk[-1].lemma_,
                             "POS": chunk[-1].pos_,
                             "TAG": chunk[-1].tag_,
                             }
                    retokenizer.merge(chunk, attrs=attrs)
                except Exception as e:
                    print(e)
        return doc

    @staticmethod
    def merge_np_of_np(doc):
        of_merge_index = []
        for of_token in doc:
            if of_token.text == "of" and of_token.head.pos_ == "NOUN":
                flag = False
                third_token_list = [child for child in of_token.children]
                if len(third_token_list) != 0:
                    third_token = third_token_list[0]
                else:
                    third_token = of_token
                if len(of_merge_index) > 0:
                    for index_pair in of_merge_index:
                        if of_token.head.left_edge.i > index_pair[0] and of_token.head.left_edge.i < index_pair[
                            1] and third_token.right_edge.i + 1 > index_pair[1]:
                            of_merge_index.append(
                                (index_pair[0], third_token.right_edge.i + 1, of_token.head.tag_, of_token.head.pos_))
                            of_merge_index.remove(index_pair)
                        if of_token.head.left_edge.i < index_pair[0] and third_token.right_edge.i + 1 > index_pair[1]:
                            of_merge_index.append((of_token.head.left_edge.i, third_token.right_edge.i + 1,
                                                   of_token.head.tag_, of_token.head.pos_))
                            of_merge_index.remove(index_pair)
                else:
                    of_merge_index.append((of_token.head.left_edge.i, third_token.right_edge.i + 1, of_token.head.tag_,
                                           of_token.head.pos_))
        try:
            with doc.retokenize() as retokenizer:
                for start, end, tag, pos in of_merge_index:
                    span = doc[start: end]
                    attrs = {"LEMMA": span.lemma_.replace("  ", " "),
                             "POS": pos,
                             "TAG": tag,
                             }
                    retokenizer.merge(span, attrs=attrs)
        except Exception as e:
            print(e)
        return doc

    @staticmethod
    def merge_clausadvmode_as_condition(doc):
        """
        merge the if condition or Attributive clause into one element to simplify
        :param doc:
        :return:
        """
        # todo: need complete
        # for chunk in doc.noun_chunks:
        #     with doc.retokenize() as retokenizer:
        #         attrs = {"LEMMA": chunk[:-1].text + " " + chunk[-1].lemma_,
        #                  "POS": chunk[-1].pos_,
        #                  "TAG": chunk[-1].tag_,
        #                  }
        #         retokenizer.merge(chunk, attrs=attrs)
        return doc

    @staticmethod
    def get_parallel_predicates(doc, main_predicate_token):
        """
        the input is the sentence Spacy Doc
        :param doc:
        :return: Span noun chunk for the subject, None for not found.
        """
        # todo:这个函数还不能覆盖所有的情况
        parallel_predicate_tokens = []
        for token in doc:
            if token.dep_ == "conj" and token.head == main_predicate_token:
                parallel_predicate_tokens.append(token)

        return parallel_predicate_tokens

    @staticmethod
    def get_object_for_preposition(doc, preposition):

        """
        为一个介词获取他的对象
        :param doc:
        :return: Span noun chunk for the subject, None for not found.
        """
        # todo:这个函数还不能覆盖所有的情况,需要进一步修改,目前只是简单返回介词后面链接的名词
        for token in doc:
            if (token.pos_ == "NOUN" or token.pos_ == "PROPN") and token.head == preposition:
                return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token)
        return None

    @staticmethod
    def get_prep_object_for_preposition(doc, predicate):
        for token in doc:
            if (
                    token.pos_ == "NOUN" or token.pos_ == "PROPN") and token.head.dep_ == "prep" and token.head.i > predicate.i:
                if token.head.head == predicate:
                    return DependencyTreeUtil.get_object_for_preposition(doc, token.head)

    @staticmethod
    def get_object_for_verb(doc, predicate):
        for token in doc:
            if (token.pos_ == "NOUN" or token.pos_ == "PROPN") and (token.head == predicate and (
                    token.dep_ == 'dobj' or token.dep_ == 'acomp' or token.dep_ == 'pobj')) or (
                    token.head is not None and token.head.head == predicate and token.dep_ == 'pobj'):
                return token

    @staticmethod
    def get_object_of_preposition_by(doc):
        """
        为被动语态找间接宾语，如名词的头结点是by
        :param doc:
        :return:
        """
        if 'by' in doc.text:
            for token in doc:
                if (token.pos_ == "NOUN" or token.pos_ == "PROPN") and token.head.text == 'by':
                    return DependencyTreeUtil.get_subtree_span_from_one_token_obj(doc, token)
        return None

    @staticmethod
    def clause_extraction(sent_doc, predicate):
        doc_list = []
        for token in sent_doc:
            if token.dep_ == 'relcl' and token.head.tag_.startswith('NN'):
                clause_span = DependencyTreeUtil.get_subtree_span_from_one_token_obj(sent_doc, token)
                subject_sent = DependencyTreeUtil.get_subtree_span_from_one_token_obj_for_vbn_and_root(sent_doc, token,
                                                                                                       predicate)
                # if 'that' in clause_span.text:
                #
                # doc_list.append(clause_span)
                # doc_list.append(subject_sent)

    @classmethod
    def get_can_be_string(cls, doc, predicate):
        aux = None
        auxpass_text = "be"
        neg_text = ""
        # if
        for token in doc:
            if token.tag_ == "MD" and token.head == predicate:
                aux = token.lemma_
            if token.tag_ == "RB" and token.dep_ == "neg" and token.head == predicate:
                neg_text = token.lemma_
            # if token.pos_ == "VERB" and token.lemma == "be" and token.head == predicate:
            #     auxpass = token
        if aux == None and doc[len(doc) - 1].text != predicate.text and doc[len(doc) - 2].text != predicate.text:
            return None
        if neg_text == "":
            return aux + " " + auxpass_text + " " + predicate.text if aux else predicate.text
        else:
            return aux + " " + neg_text + " " + auxpass_text + " " + predicate.text if aux else predicate.text



import traceback

import numpy as np
from gensim.models import KeyedVectors
from nltk import WordNetLemmatizer
from spacy.lang.en import LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES
from spacy.lemmatizer import Lemmatizer

from sekg.constant.constant import PropertyConstant, DomainConstant, WikiDataConstance
from sekg.graph.exporter.graph_data import GraphData
from sekg.text.extractor.domain_entity.nlp_util import SpacyNLPFactory
from sekg.text.extractor.domain_entity.relation_detection import RelationType
from sekg.wiki.WikiDataItem import WikiDataItem
from sekg.wiki.fusion.base import GenericKGFusion
from sekg.wiki.search_domain_wiki.wikidata_searcher import AsyncWikiSearcher
from sekg.wiki.wiki_util import WikiDataPropertyTable

"""
For each noun concept C_A, we use the following process to extend relevant Wikidata concepts and relations. First,
we search Wikidata concepts with the name and alias of C_A and collect the top 10 matched concepts for each search as the 
candidate Wikidata concepts CandWC. Second, we remove those concepts that share no names or aliases with C_A from CandWC. Third, we determine the Wikidata concept that C_A can be linked to in the
following way: 1) if there are concepts in CandW_C whose topic relevance is higher than 0.95, link C_A to the candidate concept with the highest topic relevance; 2) otherwise, if there is a concept CW
that has the highest topic relevance in CandW_C and the context similarity betweenC_A andCW is higher than 0.85, linkC_A toCW ; 3)
otherwise, if there are concepts in CandW_C whose topic relevance is higher than 0.8 and whose context similarity with C_A is higher
than 0.9, choose the conceptCW that has the highest topic relevance among them and link C_A to CW
"""


class WeightedScoreBasedGenericKGFusion(GenericKGFusion):
    INVALID_TEXTS = {"scientific article", "wikimedia template", "wikimedia list article", "wikipedia template",
                     "wikibase wikis", "wikimedia", "wikibase", "wikidata", "family", "singer", "music", "movie",
                     "book"}
    INVALID_SUBCLASS_ITEM_ID = set(["Q11424",  # film
                                    "Q15138389",  # wiki
                                    "Q7187",  # gene
                                    ])

    DEFAULT_FILTER_CONTEXT_SCORE = 0.8
    DEFAULT_FILTER_TOPIC_SCORE = 0.9

    DEFAULT_ACCEPTABLE_TOPIC_SCORE = 0.95
    DEFAULT_ACCEPTABLE_CONTEXT_SCORE = 0.85

    DEFAULT_PROXY_SERVER = "http://127.0.0.1:1080"

    DEFAULT_CONTEXT_SCORE_WEIGHT = 0.6
    DEFAULT_TOPIC_SCORE_WEIGHT = 0.4
    DEFAULT_FILTER_SUM_SCORE = 0.8

    def __init__(self, context_score_weight=DEFAULT_CONTEXT_SCORE_WEIGHT,
                 topic_score_weight=DEFAULT_TOPIC_SCORE_WEIGHT,
                 filter_sum_score=DEFAULT_FILTER_SUM_SCORE,
                 proxy_server=DEFAULT_PROXY_SERVER):
        # todo: fix this, remove this, use other lemmatizer way
        self.lemmatizer = Lemmatizer(LEMMA_INDEX, LEMMA_EXC, LEMMA_RULES)

        self.fetcher = AsyncWikiSearcher(proxy_server)
        self.graph_data = GraphData()
        self.wikidata_property_table = WikiDataPropertyTable.get_instance()
        self.embedding = {}
        self.context_score_weight = context_score_weight
        self.topic_score_weight = topic_score_weight
        self.filter_sum_score = filter_sum_score
        self.NLP = SpacyNLPFactory.create_simple_nlp_pipeline()

    def init_wd_from_cache(self, title_save_path=None, item_save_path=None):
        self.fetcher.init_from_cache(title_save_path=title_save_path, item_save_path=item_save_path)
        print("Init from cache...")

    def export_wd_cache(self, title_save_path, item_save_path):

        self.fetcher.save(item_save_path=item_save_path, title_save_path=title_save_path)

    def load_word_embedding(self, emb_path):
        wv = KeyedVectors.load(emb_path)
        self.embedding = {k: wv[k] for k in wv.vocab.keys()}

    def init_graph_data(self, graph_data_path):
        self.graph_data = GraphData.load(graph_data_path)

    def fetch_wikidata_by_name(self, terms, title_save_path=None, item_save_path=None):
        """
                search with some terms and find the candidate wikidata item list for the term,
                 and cache all the possible wikidata item for the item.
                 eg. for term: "apple", we will search it in wikidata.org by API and get the returned
                 search result list(maybe 10 result). the search result for keywords will be cached.
                 And we we retrieve all 10 candidate wikidata item info.

                :param item_save_path: the wikidata item info cache path
                :param title_save_path:  the search result by title saving path
                :param terms: a list of str or a set of str standing for concepts.
                :return:
                """
        self.fetcher.init_from_cache(title_save_path=title_save_path, item_save_path=item_save_path)
        terms = {self.lemmatizer.noun(term)[0].lower() for term in terms}
        print("need to fetch %r term wiki titles, %r are already cache, actual %r need to fetch" % (
            len(terms), len(self.fetcher.title_cache.keys() & terms),
            len(terms) - len(self.fetcher.title_cache.keys() & terms)))

        term_titles = self.fetcher.search_title(terms)
        if title_save_path is not None:
            self.fetcher.save(title_save_path=title_save_path)

        ids = self.get_valid_wikidata_item(term_titles)
        term_wikiitems = self.fetch_wikidata_by_id(ids, item_save_path)
        return term_titles, term_wikiitems

    @staticmethod
    def is_need_to_fetch_wikidata_item(item):
        INVALID_TEXTS = ["scientific article", "wikimedia template", "wikimedia list article", "wikipedia template",
                         "wikibase wikis", "wikimedia"]

        snippet = item["snippet"].lower()
        for invalid_text in INVALID_TEXTS:
            if invalid_text in snippet:
                return False

        return True

    @staticmethod
    def get_valid_wikidata_item(term_titles):
        """
        some search results for wikidata are not need to search, for example, the item has "scientific article" in description.
        :param term_titles:
        :return:
        """
        valid_wikidata_ids = set([])

        for v in term_titles.values():
            for item in v:
                if WeightedScoreBasedGenericKGFusion.is_need_to_fetch_wikidata_item(item) == False:
                    continue
                valid_wikidata_ids.add(item["title"])

        return valid_wikidata_ids

    def fetch_wikidata_by_id(self, ids, item_save_path=None):

        print("need to fetch wikidata items num=%r, %r are already cache, actual %r need to fetch" % (
            len(ids), len(self.fetcher.item_cache.keys() & ids),
            len(ids) - len(self.fetcher.item_cache.keys() & ids)))

        term_wikiitems = self.fetcher.fetch_item(ids)
        if item_save_path is not None:
            self.fetcher.save(item_save_path=item_save_path)
        return term_wikiitems

    def compute_topic_vector(self):
        topic_words = []
        for node_id in self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM):
            try:
                node_json = self.graph_data.get_node_info_dict(node_id=node_id)
                if not node_json:
                    continue
                node_properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
                lemma = node_properties[PropertyConstant.LEMMA]
                aliases = node_properties.get(PropertyConstant.ALIAS, [])
                aliases_en = node_properties.get("aliases_en", [])
                description_en = node_properties.get("descriptions_en", "")
                name = node_properties.get("name", "")
                topic_words.append(lemma)
                topic_words.extend(aliases)
                topic_words.extend(aliases_en)
                topic_words.append(description_en)
                topic_words.append(name)
            except:
                traceback.print_exc()
        topic_text = " ".join(topic_words).lower()

        if len(topic_text) == 0:
            return None
        words = [w for w in topic_text.split() if w]
        if len(words) == 0:
            return None
        vec_des = sum([self.embedding.get(w, np.zeros([100])) for w in words]) / len(words)

        return vec_des

    def compute_wikidata_vector(self, wikidata_item, term_wikiitems, domain_term_name=""):
        """
        if given the domain term name, the name containing in domain term will be exclude from the text
        :param wikidata_item:
        :param term_wikiitems:
        :param domain_term_name:
        :return:
        """
        relation_text = self.generate_relations_text(wikidata_item, term_wikiitems)
        description = wikidata_item.get_en_description()
        en_name = wikidata_item.get_en_name()
        en_aliases = wikidata_item.get_en_aliases()

        description = " ".join([en_name, " ".join(en_aliases), description, relation_text])

        # words = list(set(
        #     [token.lemma_.lower() for token in self.NLP(description) if
        #      token.is_digit == False and token.is_stop == False]))
        words = [token.lemma_.lower() for token in self.NLP(description) if
                 token.is_digit == False and token.is_stop == False and token.is_punct == False]
        # todo: fix this name

        removal_words = set(domain_term_name.lower().split())
        words = [w for w in words if w not in removal_words]

        if len(words) == 0:
            return None
        # todo: the size of vector should be adjust
        vec_des = sum([self.embedding.get(w, np.zeros([100])) for w in words]) / len(words)

        return vec_des

    def __score_topic(self, topic_vector, wikidata_item, term_wikiitems, domain_term_name=""):

        wikidata_vector = self.compute_wikidata_vector(wikidata_item, term_wikiitems, domain_term_name=domain_term_name)
        return self.compute_sim_for_two_vectors(wikidata_vector, topic_vector)

    def __score_context(self, node_json, wikidata_item, term_wikiitems):
        relation_text = self.generate_relations_text(wikidata_item, term_wikiitems)
        description = wikidata_item.get_en_description()
        en_name = wikidata_item.get_en_name()
        en_aliases = wikidata_item.get_en_aliases()

        description = " ".join([en_name, " ".join(en_aliases), description, relation_text])

        domain_term_name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES][PropertyConstant.LEMMA]

        name = self.get_compare_name_for_domain_term(node_json)

        removal_words = set(domain_term_name.lower().split())

        if len(description) == 0 or len(name) == 0:
            return 0
        # words = list(set(
        #     [token.lemma_.lower() for token in self.NLP(description) if
        #      token.is_digit == False and token.is_stop == False]))
        words = [token.lemma_.lower() for token in self.NLP(description) if
                 token.is_digit == False and token.is_stop == False and token.is_punct == False]
        words = [w for w in words if w not in removal_words]

        if len(words) == 0:
            return 0
        vec_des = sum([self.embedding.get(w, np.zeros([100])) for w in words]) / len(words)
        # name_words = list(
        #     set([token.lemma_.lower() for token in self.NLP(name) if
        #          token.is_digit == False and token.is_stop == False]))
        name_words = [token.lemma_.lower() for token in self.NLP(name) if
                      token.is_digit == False and token.is_stop == False]

        if len(name_words) == 0:
            return 0
        vec_term = sum([self.embedding.get(w, np.zeros([100])) for w in name_words]) / len(name_words)

        return self.compute_sim_for_two_vectors(vec_des, vec_term)

    def compute_sim_for_two_vectors(self, vec_des, vec_term):
        norm_des = np.linalg.norm(vec_des)
        norm_term = np.linalg.norm(vec_term)
        if norm_des == 0 or norm_term == 0:
            return 0
        return 0.5 + vec_des.dot(vec_term) / (norm_des * norm_term) / 2

    def get_compare_name_for_domain_term(self, node_json):
        domain_term_id = node_json[GraphData.DEFAULT_KEY_NODE_ID]

        name = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES].get(PropertyConstant.LEMMA, "")
        aliases = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES].get(PropertyConstant.ALIAS, [])
        aliases_en = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES].get("aliases_en", [])

        other_names = [name]
        other_names.extend(aliases)
        other_names.extend(aliases_en)

        out_relations = self.graph_data.get_all_out_relations(node_id=domain_term_id)
        in_relations = self.graph_data.get_all_in_relations(node_id=domain_term_id)
        domain_term_node_ids = self.graph_data.label_to_ids_map[DomainConstant.LABEL_DOMAIN_TERM]
        id_set = set([])
        for (start_id, r, end_id) in out_relations:
            if end_id in domain_term_node_ids:
                id_set.add(end_id)
        for (start_id, r, end_id) in in_relations:
            if start_id in domain_term_node_ids:
                id_set.add(start_id)
        id_set.add(domain_term_id)
        for id in id_set:
            temp_node_json = self.graph_data.get_node_info_dict(node_id=id)
            other_names.append(temp_node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES].get(PropertyConstant.LEMMA, ""))
        name = " ".join(other_names)
        return name

    def add_wikidata_item(self, item: WikiDataItem):
        """
        add a new term to graph data
        :param term: the term added to GraphData
        :return: the node_id fo the added term node
        """
        ori_node_json = self.graph_data.find_one_node_by_property(WikiDataConstance.PRIMARY_PROPERTY_NAME,
                                                                  item.wd_item_id)
        if ori_node_json:
            # print(ori_node_json)
            # print('no new wiki node!! node %d has fused wiki_node %s' % (ori_node_json["id"], item.wd_item_id))
            return ori_node_json["id"]
        # print("add new wikinode %s" % (item.wd_item_id))
        node_labels = [WikiDataConstance.LABEL_WIKIDATA]
        node_properties = {
            WikiDataConstance.PRIMARY_PROPERTY_NAME: item.wd_item_id,
            WikiDataConstance.NAME: item.get_en_name(),
            PropertyConstant.ALIAS: set(item.get_en_aliases()),
        }
        item.get_relation_property_name_list()
        relation_property_set = set(item.relation_property_name_list)
        pure_property_set = set(item.get_non_relation_property_name_list())

        valid_property_dict = {}
        for p, v in item.data_dict.items():
            if p in relation_property_set:
                continue
            if p in pure_property_set:
                p = self.wikidata_property_table.property_id_2_name(p)
                if p == None:
                    continue
            valid_property_dict[p] = v
        wikidata_node_id = self.graph_data.add_node(node_labels=node_labels,
                                                    node_properties=dict(valid_property_dict, **node_properties),
                                                    primary_property_name=WikiDataConstance.PRIMARY_PROPERTY_NAME)
        return wikidata_node_id

    def fuse_wikidata_item(self, domain_id, item: WikiDataItem):
        """
        add WikiDataItem into to domian term node
        :domain_id: domain_term id
        :param term: the term added to GraphData
        :return: the node_id fo the added term node
        """
        node_properties = {
            WikiDataConstance.PRIMARY_PROPERTY_NAME: item.wd_item_id,
            WikiDataConstance.NAME: item.get_en_name(),
            PropertyConstant.ALIAS: set(item.get_en_aliases()),
        }
        item.get_relation_property_name_list()
        relation_property_set = set(item.relation_property_name_list)
        pure_property_set = set(item.get_non_relation_property_name_list())

        valid_property_dict = {}
        for p, v in item.data_dict.items():
            if p in relation_property_set:
                continue
            if p in pure_property_set:
                p = self.wikidata_property_table.property_id_2_name(p)
                if p == None:
                    continue
            valid_property_dict[p] = v

        self.graph_data.add_labels(WikiDataConstance.LABEL_WIKIDATA)
        self.graph_data.add_label_by_node_id(domain_id, WikiDataConstance.LABEL_WIKIDATA)
        domain_node_json = self.graph_data.get_node_info_dict(domain_id)
        domain_properties_json = domain_node_json[self.graph_data.DEFAULT_KEY_NODE_PROPERTIES]
        if domain_node_json:
            for k, v in dict(valid_property_dict, **node_properties).items():
                if k in domain_properties_json:
                    if v == domain_properties_json[k]:
                        pass
                    else:
                        if type(v) == set:
                            domain_properties_json[k] = v.union(domain_properties_json[k])
                        if type(v) == list:
                            domain_properties_json[k].extend(v)
                        if type(v) == str:
                            domain_properties_json[k + "_wiki"] = v
                else:
                    domain_properties_json[k] = v
        # print("fuse node %d and wiki node %s" % (domain_id, item.wd_item_id))
        # print("&" * 10)
        # print(self.graph_data.get_node_info_dict(domain_id))
        return domain_id

    def init_lemma(self):
        lemmatizer = WordNetLemmatizer()
        self.graph_data.create_index_on_property(PropertyConstant.LEMMA)

        for node_id in self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM):
            node_json = self.graph_data.get_node_info_dict(node_id=node_id)

            if not node_json:
                continue

            node_properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]

            lemma_domain_term_name = node_properties[DomainConstant.PRIMARY_PROPERTY_NAME]
            lemma_domain_term_name = lemma_domain_term_name.lower()
            lemma_domain_term_name = lemmatizer.lemmatize(lemma_domain_term_name, "n")
            node_properties[PropertyConstant.LEMMA] = lemma_domain_term_name

        self.graph_data.refresh_indexer()

    def fuse(self, ):
        # todo: this fusion has dependency on GraphData, not good.
        self.graph_data.create_index_on_property(WikiDataConstance.PRIMARY_PROPERTY_NAME)
        self.init_lemma()

        term_titles = self.fetcher.get_title_cache()
        # todo: by calling the method not access the field
        term_wikiitems = self.fetcher.get_item_cache()

        id_item = {}
        record = []
        topic_vector = self.compute_topic_vector()
        if topic_vector is None:
            print("error, topic vector is None, maybe the graph has not domain term")
            return
        for node_id in self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM):
            try:
                node_json = self.graph_data.get_node_info_dict(node_id=node_id)
                if not node_json:
                    continue
                if WikiDataConstance.LABEL_WIKIDATA not in node_json["labels"]:
                    node_properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
                    lemma = node_properties[PropertyConstant.LEMMA]
                    alias_set = node_properties[PropertyConstant.ALIAS]
                    if lemma.find(".") >= 0 or lemma.find("(") >= 0:
                        continue
                    alias_set.add(lemma)
                    alias_set.add(lemma.lower())
                    items = set()
                    for lemma in alias_set:
                        if lemma not in term_titles:
                            continue
                        titles = term_titles[lemma]
                        for title in titles:
                            item = term_wikiitems.get(title["title"], None)
                            if item is None:
                                continue
                            wk_names = set()
                            for n in item.get_en_aliases():
                                wk_names.add(n)
                                wk_names.add(n.lower())

                            wk_names.add(item.get_en_name())
                            wk_names.add(item.get_en_name().lower())

                            if item is None or lemma not in wk_names:
                                continue
                            if self.is_valid_wikidata_item(item) is False:
                                continue
                            items.add(item)

                    if len(items) == 0:
                        continue
                    items = list(items)
                    sims = [(item, self.__score_context(node_json, item, term_wikiitems),
                             self.__score_topic(topic_vector, item, term_wikiitems,
                                                domain_term_name=lemma)) for item in items]

                    combined_sims = []

                    for item, context_score, topic_score in sims:
                        combined_sims.append((item, context_score, topic_score,
                                              context_score * self.context_score_weight + topic_score * self.topic_score_weight))

                    best_link_wikidata_item, best_context_score, best_topic_score, best_sum_score = max(sims,
                                                                                                        key=lambda x: x[
                                                                                                            3])

                    is_link = False
                    if best_sum_score > self.filter_sum_score:
                        is_link = True

                    record.append({
                        "name": best_link_wikidata_item.get_en_name(),
                        "alias": best_link_wikidata_item.get_en_aliases(),
                        "description": best_link_wikidata_item.get_en_description(),
                        "wk relation text": self.generate_relations_text(best_link_wikidata_item, term_wikiitems),
                        "domain term": lemma,
                        "combine_score": best_sum_score,
                        "context_score": best_context_score,
                        "topic_score": best_topic_score,
                        "combined name": self.get_compare_name_for_domain_term(node_json),
                        "link": is_link,
                        "domain_id": node_id,
                        "wd_item_id": best_link_wikidata_item.wd_item_id
                    })

                    if best_link_wikidata_item == None:
                        continue
                    if is_link == False:
                        continue
                    # wikidata_node_id = self.add_wikidata_item(best_link_wikidata_item)
                    if is_link:
                        wikidata_node_id = self.fuse_wikidata_item(node_id, best_link_wikidata_item)
                        if wikidata_node_id == GraphData.UNASSIGNED_NODE_ID:
                            continue
                        id_item[wikidata_node_id] = best_link_wikidata_item
            except Exception:
                traceback.print_exc()
        self.build_relation_between_wikidata_node_in_graph(term_wikiitems)
        neighbours = set()
        for _id, item in id_item.items():
            for r in item.relation_property_name_list:
                end_id_set = self.get_wikidata_item_ids_by_relation(item, r)
                for e in end_id_set:
                    neighbours.add(e)
        self.graph_data.refresh_indexer()
        return neighbours, record

    def fuse_with_prefix_and_suffix(self):
        domain_ids = self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM)
        wiki_ids = self.graph_data.get_node_ids_by_label(WikiDataConstance.LABEL_WIKIDATA)
        unfuse_domain = domain_ids - wiki_ids
        unfuse_wiki = wiki_ids - domain_ids
        relations_set = set()
        domain_wiki_map = {}
        relations = []
        domain_wiki_fuse = []
        lemmatizer = WordNetLemmatizer()

        clean_wiki_name_map = {}

        for wiki_id in wiki_ids:
            wiki_node_info = self.graph_data.get_node_info_dict(wiki_id)
            wiki_node_id = wiki_node_info[GraphData.DEFAULT_KEY_NODE_ID]
            wikidata_name = wiki_node_info[GraphData.DEFAULT_KEY_NODE_PROPERTIES][WikiDataConstance.NAME]
            aliases = wiki_node_info[GraphData.DEFAULT_KEY_NODE_PROPERTIES][PropertyConstant.ALIAS]

            clean_wiki_name_map[wiki_node_id] = set()
            clean_wikidata_name = lemmatizer.lemmatize(wikidata_name.lower(), "n")

            if len(clean_wikidata_name) > 2:
                clean_wiki_name_map[wiki_node_id].add((wikidata_name, clean_wikidata_name))
            for ali in aliases:
                clean_wiki_alise = lemmatizer.lemmatize(ali.lower(), "n")

                if len(clean_wiki_alise) > 2:
                    clean_wiki_name_map[wiki_node_id].add((ali, clean_wiki_alise))
            if not clean_wiki_name_map[wiki_node_id]:
                clean_wiki_name_map.pop(wiki_node_id)

        for domain_term_id in unfuse_domain:
            domain_node_info = self.graph_data.get_node_info_dict(domain_term_id)
            domain_name = domain_node_info[self.graph_data.DEFAULT_KEY_NODE_PROPERTIES][
                DomainConstant.PRIMARY_PROPERTY_NAME]
            domain_name_list = domain_name.lower().split(" ")
            domain_name = lemmatizer.lemmatize(domain_name.lower(), "n")

            for wikidata_node_id in clean_wiki_name_map:
                for wiki_name, clean_wiki_name in clean_wiki_name_map[wikidata_node_id]:
                    if domain_name == clean_wiki_name:
                        if domain_term_id not in domain_wiki_map:
                            domain_wiki_map[domain_term_id] = []
                            domain_wiki_map[domain_term_id].append(wikidata_node_id)
                        else:
                            domain_wiki_map[domain_term_id].append(wikidata_node_id)
                        domain_wiki_fuse.append(
                            {"domain_id": domain_term_id, "domain_name": domain_name, "wiki_id": wikidata_node_id,
                             "wiki_name": wiki_name})
                        continue
                    if len(domain_name_list) > 1:
                        if domain_name.startswith(clean_wiki_name + " "):
                            relations_set.add((domain_term_id, RelationType.DERIVED_FROM.value, wikidata_node_id))
                            relations.append(
                                {"domain_id": domain_term_id, "domain_name": domain_name,
                                 "relation": RelationType.DERIVED_FROM.value, "wiki_id": wikidata_node_id,
                                 "wiki_name": wiki_name})
                        if domain_name.endswith(" " + clean_wiki_name):
                            relations_set.add((domain_term_id, RelationType.IS_A.value, wikidata_node_id))
                            relations.append(
                                {"domain_id": domain_term_id, "domain_name": domain_name,
                                 "relation": RelationType.IS_A.value, "wiki_id": wikidata_node_id,
                                 "wiki_name": wiki_name})

        print("and new domian_wiki relation %d" % (len(relations_set)))
        for item in relations_set:
            s = item[0]
            r = item[1]
            e = item[2]
            self.graph_data.add_relation(s, r, e)

        # fuse wiki_node and domain_node , they have the same name
        for domain_id in domain_wiki_map:
            wiki_id = domain_wiki_map[domain_id][0]
            self.graph_data.merge_two_nodes_by_id(main_node_id=wiki_id, merged_node_id=domain_id,
                                                  main_alias_property=PropertyConstant.ALIAS,
                                                  merged_alias_property=PropertyConstant.ALIAS)

        return domain_wiki_fuse, relations

    @staticmethod
    def get_wikidata_item_ids_by_relation(wikidata_item: WikiDataItem, r):
        id_set = set([])
        end = wikidata_item.data_dict.get(r, [])
        if type(end) == list:
            for e in end:
                id_set.add(e)
        else:
            id_set.add(end)
        return id_set

    def generate_relations_text(self, wikidata_item, term_wikiitems):
        text = []
        for r in wikidata_item.relation_property_name_list:

            relation_name = self.wikidata_property_table.property_id_2_name(r)
            if relation_name == None:
                relation_name = r
            end = wikidata_item.data_dict[r]

            if type(end) == list:
                for e_wd_item_id in end:
                    if self.is_valid_wikidata_item_id(e_wd_item_id):
                        neibour_item = term_wikiitems.get(e_wd_item_id, None)
                        if neibour_item != None:
                            text.append(neibour_item.get_en_name())
                            # if relation_name in {"subclass of", "instance of", "part of"}:
                            #     text.append(neibour_item.get_en_description())
                            text.append(neibour_item.get_en_description())

                    else:
                        text.append(end)
                text.append(relation_name)

            else:
                if self.is_valid_wikidata_item_id(end):
                    neibour_item = term_wikiitems.get(end, None)
                    if neibour_item != None:
                        text.append(neibour_item.get_en_name())
                        # if relation_name in {"subclass of", "instance of", "part of"}:
                        #     text.append(neibour_item.get_en_description())
                        text.append(neibour_item.get_en_description())
                else:
                    text.append(end)
                text.append(relation_name)

        return " ".join(text)

    def is_valid_wikidata_item_id(self, wd_item_id):
        try:

            if wd_item_id.startswith("Q") and wd_item_id[1:].isdigit():
                return True
            return False
        except:
            return False

    def get_all_neighbours_id(self, item):
        neighbours = set()
        for r in item.relation_property_name_list:
            end = item.data_dict[r]
            if type(end) == list:
                for e in end:
                    if e[0] == "Q" or e[0] == "P":
                        neighbours.add(e)
            else:
                if end[0] == "Q" or end[0] == "P":
                    neighbours.add(end)

        return neighbours

    def get_all_neighbours_id_by_item_id(self, item_id):
        neighbours = set()
        item = self.fetcher.item_cache.get(item_id, None)
        if item == None:
            return set()
        for r in item.relation_property_name_list:
            end = item.data_dict[r]
            if type(end) == list:
                for e in end:
                    if e[0] == "Q" or e[0] == "P":
                        neighbours.add(e)
            else:
                if end[0] == "Q" or end[0] == "P":
                    neighbours.add(end)

        return neighbours

    def fetch_valid_wikidata_item_neibours_from_all_term_titles(self, item_save_path):
        """
        some search results for wikidata are not need to search, for example, the item has "scientific article" in description.
        :param term_titles:
        :return:
        """
        term_titles = self.fetcher.title_cache
        valid_wikidata_ids = GenericKGFusion.get_valid_wikidata_item(term_titles)
        nerbours = set([])

        for valid_id in valid_wikidata_ids:
            nerbours.update(self.get_all_neighbours_id_by_item_id(valid_id))
        return self.fetch_wikidata_by_id(nerbours, item_save_path)

    def add_wikidata_items(self, wd_item_ids):
        term_wikiitems = self.fetcher.item_cache
        self.graph_data.refresh_indexer()
        for wd_item_id in wd_item_ids:
            if wd_item_id in term_wikiitems:
                self.add_wikidata_item(term_wikiitems[wd_item_id])
        self.build_relation_between_wikidata_node_in_graph(term_wikiitems)

    def build_relation_between_wikidata_node_in_graph(self, term_wikiitems):
        wikidata_node_ids = self.graph_data.get_node_ids_by_label(WikiDataConstance.LABEL_WIKIDATA)
        wd_item_id_2_node_id_map = {}
        node_id_2_wd_item_id_map = {}
        for node_id in wikidata_node_ids:
            wikidata_node = self.graph_data.get_node_info_dict(node_id)
            wd_item_id = wikidata_node[GraphData.DEFAULT_KEY_NODE_PROPERTIES][WikiDataConstance.PRIMARY_PROPERTY_NAME]
            wd_item_id_2_node_id_map[wd_item_id] = node_id
            node_id_2_wd_item_id_map[node_id] = wd_item_id
        for start_wd_item_id, start_node_id in wd_item_id_2_node_id_map.items():
            start_wikidata_item = term_wikiitems.get(start_wd_item_id, None)
            if start_wikidata_item == None:
                continue
            for r_id in start_wikidata_item.relation_property_name_list:
                end_wd_ids = self.get_wikidata_item_ids_by_relation(start_wikidata_item, r_id)
                relation_name = self.wikidata_property_table.property_id_2_name(r_id)
                if relation_name == None:
                    continue

                for end_wd_id in end_wd_ids:
                    end_node_id = wd_item_id_2_node_id_map.get(end_wd_id, None)
                    if end_node_id == None:
                        continue
                    if start_node_id == end_node_id:
                        continue
                    self.graph_data.add_relation(start_node_id, relation_name, end_node_id)

    def save(self, graph_data_path):
        self.graph_data.save(graph_data_path)

    def is_valid_wikidata_item(self, item):
        for text in self.INVALID_TEXTS:
            en_name = item.get_en_name().lower()
            if text in en_name:
                return False
            description = item.get_en_description()
            if text in description:
                return False
        end_wd_ids = self.get_wikidata_item_ids_by_relation(item, "P31")

        for end_wd in end_wd_ids:
            if end_wd in self.INVALID_SUBCLASS_ITEM_ID:
                return False

        return True

    def fetch_wikidata_by_name_and_cache_neibours(self, terms, title_save_path, item_save_path):
        self.fetch_wikidata_by_name(terms, item_save_path=item_save_path, title_save_path=title_save_path)
        self.fetch_valid_wikidata_item_neibours_from_all_term_titles(item_save_path=item_save_path)

    def merge_domain_nodes_fuse_same_wiki_item(self, nodes):
        for i in range(1, len(nodes)):
            print("merge %d && %d !" % (nodes[0], nodes[i]))
            self.graph_data.merge_two_nodes_by_id(nodes[0], nodes[i], PropertyConstant.ALIAS, PropertyConstant.ALIAS)
            # print(self.graph_data.get_node_info_dict(nodes[0]))



import numpy as np


class InformationRetrievalMetric:
    """
    Compute the metric for experiment. Here is the detail for all metric.
    https://blog.csdn.net/lsgqjh/article/details/79509169?utm_source=blogxgwz6
    """

    @staticmethod
    def precision(data):
        """
        compute the precision
        :param data: the data must be a tuple.   list 0,1,eg.([1,0,...],5).  In each tuple (actual result,ground truth num),ground truth num is the total ground num.
         ([1,0,...],5),
        or list of tuple eg. [([1,0,1,...],5),([1,0,...],6),([0,0,...],5)].
        1 stands for a correct answer, 0 stands for a wrong answer.

        :return: if input data is list, return the precision of this list. if the input data is list of list, return the
        average precision on all list. The second return value is a list of precision for each input.
        """
        if type(data) != list and type(data) != tuple:
            raise Exception("the input must be a tuple([0,...,1,...],int) or a iteration of list of tuple")

        if len(data) == 0:
            return 0.0, [0.0]
        if type(data) == tuple:
            if data[1] == 0:
                return 0.0, [0.0]
            else:
                precision = np.mean(data[0])
                return precision, [precision]

        if type(data) == list:
            separate_result = []
            for sub_list, total_num in data:
                if total_num == 0:
                    precision = 0
                else:
                    precision = np.mean(sub_list)

                separate_result.append(precision)
            return np.mean(separate_result), separate_result

    @staticmethod
    def recall(data):
        """
        compute the recall
        :param data: the data must be a tuple, list 0,1,eg.([1,0,...],5).  In each tuple (actual result,ground truth num),ground truth num is the total ground num.
         ([1,0,...],5),
        or list of tuple eg. [([1,0,1,...],5),([1,0,...],6),([0,0,...],5)].
        1 stands for a correct answer, 0 stands for a wrong answer.

        :return: if input data is list, return the recall of this list. if the input data is list of list, return the
        average recall on all list. The second return value is a list of precision for each input.
        """
        if type(data) != list and type(data) != tuple:
            raise Exception("the input must be a tuple([0,...,1,...],int) or a iteration of list of tuple")

        if len(data) == 0:
            return 0.0, [0.0]
        if type(data) == tuple:
            if data[1] == 0:
                return 0.0, [0.0]
            else:
                recall = np.sum(data[0]) / data[1]
                return recall, [recall]

        if type(data) == list:
            separate_result = []
            for sub_list, total_num in data:
                if total_num == 0:
                    recall = 0
                else:
                    recall = np.sum(sub_list) / total_num

                separate_result.append(recall)
            return np.mean(separate_result), separate_result

    @staticmethod
    def f1(data):
        """
        compute the f1
        :param data: the data must be a tuple, list 0,1,eg.([1,0,...],5).  In each tuple (actual result,ground truth num),ground truth num is the total ground num.                 ([1,0,...],5),
        or list of tuple eg. [([1,0,1,...],5),([1,0,...],6),([0,0,...],5)].
        1 stands for a correct answer, 0 stands for a wrong answer.

        :return: if input data is list, return the recall of this list. if the input data is list of list, return the
        average recall on all list. The second return value is a list of precision for each input.
        """
        if type(data) != list and type(data) != tuple:
            raise Exception("the input must be a tuple([0,...,1,...],int) or a iteration of list of tuple")

        if len(data) == 0:
            return 0.0, [0.0]
        if type(data) == tuple:

            precision, _ = InformationRetrievalMetric.precision(data)
            recall, _ = InformationRetrievalMetric.recall(data)

            if precision == 0 or recall == 0:
                f1 = 0.0
            else:
                f1 = 2 * precision * recall / (precision + recall)
            return f1, [f1]

        if type(data) == list:
            separate_result = []
            for single_data in data:
                precision, _ = InformationRetrievalMetric.precision(single_data)
                recall, _ = InformationRetrievalMetric.recall(single_data)

                if precision == 0 or recall == 0:
                    f1 = 0.0
                else:
                    f1 = 2 * precision * recall / (precision + recall)

                separate_result.append(f1)
            return np.mean(separate_result), separate_result

    @staticmethod
    def mrr(data):
        """
        compute the MRR
        :param data: the data must be a tuple, list 0,1,eg.([1,0,...],5).  In each tuple (actual result,ground truth num),ground truth num is the total ground num.
         ([1,0,...],5),
        or list of tuple eg. [([1,0,1,...],5),([1,0,...],6),([0,0,...],5)].
        1 stands for a correct answer, 0 stands for a wrong answer.

        :return: if input data is list, return the recall of this list. if the input data is list of list, return the
        average recall on all list. The second return value is a list of precision for each input.
        """
        if type(data) != list and type(data) != tuple:
            raise Exception("the input must be a tuple([0,...,1,...],int) or a iteration of list of tuple")

        if len(data) == 0:
            return 0.0, [0.0]
        if type(data) == tuple:
            (sub_list, total_num) = data
            sub_list = np.array(sub_list)
            if total_num == 0:
                return 0.0, [0.0]
            else:
                ranking_array = 1.0 / (np.array(list(range(len(sub_list)))) + 1)
                mr_np = sub_list * ranking_array

                mr = 0.0
                for team in mr_np:
                    if team > 0:
                        mr = team
                        break
                return mr, [mr]

        if type(data) == list:
            separate_result = []
            for (sub_list, total_num) in data:
                sub_list = np.array(sub_list)

                if total_num == 0:
                    mr = 0.0
                else:
                    ranking_array = 1.0 / (np.array(list(range(len(sub_list)))) + 1)
                    mr_np = sub_list * ranking_array

                    mr = 0.0
                    for team in mr_np:
                        if team > 0:
                            mr = team
                            break

                separate_result.append(mr)
            return np.mean(separate_result), separate_result

    @staticmethod
    def map(data):
        """
        compute the MAP
        :param data: the data must be a tuple, list 0,1,eg.([1,0,...],5).  In each tuple (actual result,ground truth num),ground truth num is the total ground num.
         ([1,0,...],5),
        or list of tuple eg. [([1,0,1,...],5),([1,0,...],6),([0,0,...],5)].
        1 stands for a correct answer, 0 stands for a wrong answer.

        :return: if input data is list, return the recall of this list. if the input data is list of list, return the
        average recall on all list. The second return value is a list of precision for each input.
        """
        if type(data) != list and type(data) != tuple:
            raise Exception("the input must be a tuple([0,...,1,...],int) or a iteration of list of tuple")

        if len(data) == 0:
            return 0.0, [0.0]
        if type(data) == tuple:
            (sub_list, total_num) = data
            sub_list = np.array(sub_list)
            if total_num == 0:
                return 0.0, [0.0]
            else:
                ranking_array = 1.0 / (np.array(list(range(len(sub_list)))) + 1)

                right_ranking_list = []
                count = 1
                for t in sub_list:
                    if t == 0:
                        right_ranking_list.append(0)
                    else:
                        right_ranking_list.append(count)
                        count += 1

                ap = np.sum(np.array(right_ranking_list) * ranking_array) / total_num
                return ap, [ap]

        if type(data) == list:
            separate_result = []
            for (sub_list, total_num) in data:
                sub_list = np.array(sub_list)

                if total_num == 0:
                    ap = 0.0
                else:
                    ranking_array = 1.0 / (np.array(list(range(len(sub_list)))) + 1)

                    right_ranking_list = []
                    count = 1
                    for t in sub_list:
                        if t == 0:
                            right_ranking_list.append(0)
                        else:
                            right_ranking_list.append(count)
                            count += 1

                    ap = np.sum(np.array(right_ranking_list) * ranking_array) / total_num

                separate_result.append(ap)
            return np.mean(separate_result), separate_result

    @staticmethod
    def fleiss_kappa(data, N, k, n):
        """
        compute fleiss_kappa
        :param data:  N*n matrix, number of people who label task  with specific answer, such as [[0, 0, 0, 0, 14],
            [0, 2, 6, 4, 2]] , there are two task ,each task have 5 optional answers , [0,4]=14 means this task has 14 people label the last answer
        :param N: number of task
        :param k: 0ptional answer number for each task
        :param n: number of experimenter
        :return: fleiss_kappa()float , the consensus rate of each question(list)
        """
        dataMat = np.mat(data, float)
        oneMat = np.ones((k, 1))
        sum = 0.0
        P0 = 0.0
        Pe = 0.0
        Pi_list = []
        Pj_list = []
        for i in range(N):
            temp = 0.0
            for j in range(k):
                sum += dataMat[i, j]
                temp += 1.0 * dataMat[i, j] ** 2
            temp -= n
            temp /= (n - 1) * n
            P0 += temp
            Pi_list.append(temp)
        P0 = 1.0 * P0 / N
        ysum = np.sum(dataMat, axis=0)
        # sum all element , not N * n , because one task may be have two or more answers
        if sum < N * n:
            raise Exception("input data is wrong!")
        for i in range(k):
            pj = ysum[0, i] / sum
            Pj_list.append(pj)
            Pe += pj ** 2
        ans = float(P0 - Pe) / (1 - Pe)
        return ans[0, 0], Pi_list



from unittest import TestCase

from sekg.util.code import CodeElementNameUtil


class TestCodeElementNameUtil(TestCase):
    def test_uncamelize(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "java.util. List"),
            ("java.util.ArrayList", "java.util. Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),
            ("Word2Vec", "Word2Vec"),
            ("arrayList", "array List"),
            ("HTML5", "HTML5"),
            ("0xB8CFE5","0xB8CFE5"),
            ("AWT's threading model", "AWT's threading model"),
            ("32-bit integer","32-bit integer")

        ]
        for name, right_name in name_list:
            print(right_name, util.uncamelize(name))
            # self.assertEqual(right_name, util.uncamelize(name))

    def test_uncamelize_by_stemming(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "java.util. List"),
            ("java.util.ArrayList", "java.util. Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),
            ("HTML5", "HTML5"),
            ("word2vec", "word2vec"),
            ("Word2Vec", "Word2Vec"),
            ("Word2Vec2", "Word2Vec"),
            ("3D", "3D"),
            ("three TreePaths", "three Tree Paths"),
            ("0xB8CFE5","0xB8CFE5"),
            ("AWT 's threading model","AWT 's threading model"),
            ("32-bit integer", "32-bit integer"),
        ]
        for name, right_name in name_list:
            print(right_name, util.uncamelize_by_stemming(name))
            #self.assertEqual(right_name, util.uncamelize_by_stemming(name))

    def test_uncamelize_from_simple_name(self):
        util = CodeElementNameUtil()
        name_list = [
            ("java.util.List", "List"),
            ("java.util.ArrayList", "Array List"),
            ("ArrayList", "Array List"),
            ("FPath1", "F Path1"),
            ("FilePath1", "File Path1"),
        ]
        for name, right_name in name_list:
            self.assertEqual(right_name, util.uncamelize_from_simple_name(name))

    def test_generate_aliases(self):
        util = CodeElementNameUtil()
        name_list = [
            "java.util.List",
            "java.util.ArrayList",
            "ArrayList",
            "FPath1",
            "FilePath1",
            "HTML5",
            "word2vec",
            "NEWARRAY",
            "AWT 's threading model"
        ]
        for name in name_list:
            print("ori: ", name)
            print("aliase:", util.generate_aliases(name))



from unittest import TestCase

from sekg.wiki.search_domain_wiki.wikidata_searcher import AsyncWikiSearcher


class TestWikiSearcher(TestCase):

    def test_wiki_searcher(self):
        test_set = set()
        test_set.add("Peer alarm")
        searcher = AsyncWikiSearcher(proxy_server="http://127.0.0.1:1080")
        ids = ['Q1670525']
        titles = ['Shoutbox']
        searcher.search_title(titles)
        print(searcher.get_title_cache())
        title_path = "title.bin"
        searcher.save(title_save_path=title_path)
        searcher.update(title_path, type=searcher.TYPE_TITLE)

        searcher.fetch_item(ids)
        print(searcher.get_item_cache())
        item_path = "item.bin"
        searcher.save(item_save_path=item_path)
        searcher.update(item_path, type=searcher.TYPE_ITEM)

        searcher.fetch_item_neighbor(["Q21028"])
        print(searcher.get_item_cache())

        searcher.fetch_wikipedia_context(titles)
        print(searcher.get_wikipedia_cache())
        wiki_path = "wikipedia_context.bin"
        searcher.save(wikipedia_content_save_path=wiki_path)
        searcher.update(wiki_path, type=searcher.TYPE_WIKIPEDIA)

        searcher.fetch_wikipedia_context_for_wikidata(ids)
        print(searcher.get_wikidata_wikipedia_context_cache())

        searcher.fetch_wikipedia_context_html(titles)
        print(searcher.get_wikipedia_content_html())

        searcher.fetch_wikidata_item_by_wikipedia_title(["Prototype-based_programming"])
        print(searcher.item_cache)



# # -*-coding:utf-8-*-
import os
import urllib
from urllib.parse import quote
import time
import gensim
import requests
from bs4 import BeautifulSoup
import zipfile
from xml.etree.ElementTree import parse
from sekg.util.annotation import catch_exception
from pathlib import Path
from sekg.so.SOTagItem import SOTagItem
from sekg.so.SOTagItemCollection import SOTagItemCollection


class SOTagSearcher():
    maxTryNum = 5
    user_agent = {"user-agent": "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; SV1; .NET CLR 1.1.4322)"}

    def __init__(self, so_tag_item_collection_path=None):
        self.int_tags = self.__int_tags()
        self.synonyms_data = {}
        self.init_all_synonyms()
        if so_tag_item_collection_path:
            self.so_tag_item_collection: SOTagItemCollection = SOTagItemCollection.load(so_tag_item_collection_path)
        else:
            self.so_tag_item_collection = SOTagItemCollection()

    @catch_exception
    def __int_tags(self):
        so_path = os.path.dirname(os.path.abspath(__file__))
        tag_path = str(Path(so_path) / "stackoverflow.com-Tags.zip")
        if os.path.exists(tag_path):
            with zipfile.ZipFile(tag_path, 'r') as z:
                f = z.open('Tags.xml')
            doc = parse(f)
            root = doc.getroot()
            tag_data = []
            for child in root:
                tag_data.append(child.attrib)
            tag_dic = {}
            for item in tag_data:
                tag_dic[item["TagName"].lower()] = item
            return tag_dic
        else:
            return {}

    def parse_synonyms(self, html):
        if html:
            soup = BeautifulSoup(html, 'html.parser', from_encoding='utf-8')
            div = soup.find('div', class_="mainbar")
            syn_set = set()
            if div:
                a_list = div.find_all('a', class_="post-tag")
                for item in a_list:
                    syn_set.add(item.text)
            return syn_set
        else:
            return set()

    def parse_tag_info(self, html):
        short_text = ""
        long_text = ""
        info_html = ""
        if html:
            soup = BeautifulSoup(html, 'html.parser', from_encoding='utf-8')
            text_div = soup.find('div', class_='post-text')
            short_div = text_div.find('div', class_="welovestackoverflow")
            if short_div:
                p = short_div.find('p')
                short_text = p.text
                info_html += str(short_div)
            long_text = ""
            while True:
                item = short_div.find_next_sibling()
                item_current = str(item)
                if item_current.startswith("<p>"):
                    long_text += item.text + "\n"
                    info_html += str(item)
                    short_div = item
                else:
                    break
            return short_text, long_text, info_html
        else:
            return short_text, long_text, info_html

    def download_html(self, url):
        try:
            html = requests.get(url, headers=self.user_agent).content
            if html:
                return html
            else:
                return ""
        except urllib.error.URLError as e:
            if hasattr(e, "code"):
                print(e.code)
            if hasattr(e, "reason"):
                print(e.reason)
            time.sleep(5)
            return ""

    def __fetch_synonyms_for_one_tag(self, tag):
        try:
            url = 'https://stackoverflow.com/tags/' + quote(tag) + "/synonyms"
            # print("url now: ", url)
            for tries in range(self.maxTryNum):
                html = self.download_html(url)
                if not html:
                    print("Dear Host,I can not find the web page for tag", tag)
                else:
                    parse_result = self.parse_synonyms(html)
                    if parse_result:
                        synonyms = parse_result - {tag}
                        return synonyms
            return {}
        except urllib.error.URLError as e:
            if hasattr(e, "code"):
                print(e.code)
            if hasattr(e, "reason"):
                print(e.reason)
            time.sleep(5)
            return {}
        except Exception as e:
            print("exception:" + str(e))
            time.sleep(5)
            return {}

    def __fetch_tag_info_for_one_tag(self, tag):
        try:
            url = 'https://stackoverflow.com/tags/' + quote(tag) + "/info"
            # print("url now: ", url)
            for tries in range(self.maxTryNum):
                html = self.download_html(url)
                if not html:
                    print("Dear Host,I can not find the web page for tag", tag)
                else:
                    short_text, long_text, info_html = self.parse_tag_info(html)
                    if short_text or long_text:
                        return short_text, long_text, info_html
            return "", "", ""
        except urllib.error.URLError as e:
            if hasattr(e, "code"):
                print(e.code)
            if hasattr(e, "reason"):
                print(e.reason)
            time.sleep(5)
            return "", "", ""
        except Exception as e:
            print("exception:" + str(e))
            time.sleep(5)
            return "", "", ""

    def get_tag_item_for_one_tag(self, tag):
        tag = tag.lower()
        if tag in self.so_tag_item_collection.get_all_tag_name():
            so_tag_item: SOTagItem = self.so_tag_item_collection.get_so_tag_item(tag)
            if so_tag_item.is_valid():
                print("info of {} is completed".format(tag))
                return so_tag_item
            else:
                if not so_tag_item.get_long_description() and not so_tag_item.get_short_description():
                    short_text, long_text, info_html = self.__fetch_tag_info_for_one_tag(tag)
                    so_tag_item.update_short_description(short_text)
                    so_tag_item.update_long_description(long_text)
                    so_tag_item.update_info_html(info_html)
                if not so_tag_item.get_tag_synonyms():
                    # synonyms = self.__fetch_synonyms_for_one_tag(tag)
                    # so_tag_item.update_tag_synonyms(synonyms)
                    # if tag not self.synonyms_data, it maybe no synonyms originally
                    if tag in self.synonyms_data:
                        so_tag_item.update_tag_synonyms(self.synonyms_data[tag])
                self.so_tag_item_collection.add_so_tag_item(so_tag_item)
                return so_tag_item
        elif tag in self.int_tags.keys():
            short_text, long_text, info_html = self.__fetch_tag_info_for_one_tag(tag)
            synonyms = self.__fetch_synonyms_for_one_tag(tag)
            item = self.int_tags[tag]
            so_tag_item = SOTagItem(tag_info_dic=item)
            so_tag_item.update_short_description(short_text)
            so_tag_item.update_long_description(long_text)
            so_tag_item.update_info_html(info_html)
            so_tag_item.update_tag_synonyms(synonyms)
            self.so_tag_item_collection.add_so_tag_item(so_tag_item)
            return so_tag_item
        else:
            return None

    def get_tag_item(self, tags):
        so_tag_collection = SOTagItemCollection()
        for tag in tags:
            tag = tag.lower()
            tag_item = self.get_tag_item_for_one_tag(tag)
            if tag_item:
                so_tag_collection.add_so_tag_item(tag_item)
        return so_tag_collection

    def init_all_synonyms(self):
        try:
            page = 1
            while True:
                url_tag = "http://api.stackexchange.com/2.2/tags/synonyms?page={}&pagesize=100&order=desc&sort=creation&site=stackoverflow".format(
                    str(page))
                response = requests.get(url_tag, headers=self.user_agent).json()
                if response["items"]:
                    for item in response["items"]:
                        tag_from = item["from_tag"]
                        tag_to = item["to_tag"]
                        if tag_to not in self.synonyms_data.keys():
                            self.synonyms_data[tag_to] = {tag_from}
                        else:
                            self.synonyms_data[tag_to].add(tag_from)
                else:
                    break
                page += 1
        except Exception as e:
            print("exception:" + str(e))
            time.sleep(30)

    def run(self):
        # for tag_name in self.int_tags:
        #     self.get_tag_item_for_one_tag(tag_name)
        # return self.so_tag_item_collection

        # we can use api get all synonyms, and the number of tags which have synonyms is about 2500, use api is more faster
        self.init_all_synonyms()
        valid_tag = set()
        for tag, so_tag_item in self.so_tag_item_collection.name2sotag:
            if tag in self.synonyms_data.keys():
                so_tag_item.update_tag_synonyms()
                self.so_tag_item_collection.add_so_tag_item(so_tag_item)
            if so_tag_item.get_long_description or so_tag_item.get_short_description:
                valid_tag.add(tag)

        need_tag = set(self.int_tags.keys()) - valid_tag
        for tag in need_tag:
            short_text, long_text, info_html = self.__fetch_tag_info_for_one_tag(tag)
            item = self.int_tags[tag]
            so_tag_item = SOTagItem(tag_info_dic=item)
            so_tag_item.update_short_description(short_text)
            so_tag_item.update_long_description(long_text)
            so_tag_item.update_info_html(info_html)
            if tag in self.synonyms_data:
                so_tag_item.update_tag_synonyms(self.synonyms_data[tag])
            self.so_tag_item_collection.add_so_tag_item(so_tag_item)

    def get_tag_info_cache(self):
        return self.so_tag_item_collection

    def save(self, path):
        self.so_tag_item_collection.save(path)

    def init(self, path):
        if path and os.path.exists(path):
            self.so_tag_item_collection: SOTagItemCollection = SOTagItemCollection.load(path)

    def __repr__(self):
        return "<AsyncWikiSearcher tags_count=%r>" % (len(self.so_tag_item_collection.size()))


if __name__ == '__main__':
    searcher = SOTagSearcher()
    print(len(searcher.synonyms_data))



import abc


class EntityLinkResult:
    def __init__(self, mention, node_id, node_name=None, score=1.0, **extra_info):
        self.mention = mention
        self.node_id = node_id
        self.node_name = node_name
        self.score = score

        self.extra_info = extra_info

    def __repr__(self):
        return "<EntityLinkResult mention=%r id=%r name=%r score=%r>" % (
            self.mention, self.node_id, self.node_name, self.score)


class EntryPointLinker:
    @abc.abstractmethod
    def link(self, query, **config):
        pass

    @abc.abstractmethod
    def link_n(self, query, **config):
        pass



from nltk import WordNetLemmatizer
from sekg.text.extractor.domain_entity.word_util import WordUtil
from spacy.tokens.doc import Doc


class SoftwareTextPOSFixer:
    __se_text_fixer = 'se_text_fixer'
    lemmatizer = WordNetLemmatizer()
    verb_set = {'appends', 'reads'}
    @staticmethod
    def fixer_for_pos(doc: Doc):
        if doc is None or len(doc) == 0:
            return doc
        first_token = doc[0]

        attrs = None
        if first_token.text.lower() == "return":
            attrs = {"LEMMA": "return", "TAG": "VB", "POS": "VERB"}
        if first_token.text.lower() == "returns":
            attrs = {"LEMMA": "return", "TAG": "VBZ", "POS": "VERB"}
        # if first_token.text.lower() == "appends":
        #     attrs = {"LEMMA": "append", "TAG": "VBZ", "POS": "VERB"}

        if attrs != None:
            with doc.retokenize() as retokenizer:
                retokenizer.merge(doc[0:1], attrs=attrs)
            return doc
        # todo: 考虑动词被动？
        if WordUtil.couldBeVerb(doc[0].text) and doc[0].pos_ == "ADV":
            lemma = SoftwareTextPOSFixer.lemmatizer.lemmatize(first_token.text, "v")
            if lemma == first_token.text.lower():
                tag = "VB"
            else:
                tag = "VBZ"

            attrs = {"LEMMA": lemma, "POS": "VERB",
                     "TAG": tag}
            with doc.retokenize() as retokenizer:
                retokenizer.merge(doc[0:1], attrs=attrs)
            return doc

        # 被动形式
        # if WordUtil.couldBeVerb(doc[0].text) and doc[0].tag_ == "VBD":
        #     lemma = SoftwareTextPOSFixer.lemmatizer.lemmatize(first_token.text, "v")
        #     if lemma == first_token.text.lower():
        #         tag = "VB"
        #     else:
        #         tag = "VBZ"
        #
        #     attrs = {"LEMMA": lemma, "POS": "VERB",
        #              "TAG": tag}
        #     with doc.retokenize() as retokenizer:
        #         retokenizer.merge(doc[0:1], attrs=attrs)
        #     return doc

        return doc



import networkx as nx
from gensim.models.keyedvectors import Word2VecKeyedVectors
from nltk import WordNetLemmatizer

import numpy as np
from gensim import corpora, models
from fuzzywuzzy import fuzz

from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.util.name_searcher import KGNameSearcher
from sekg.ir.models.base import DocumentSimModel
from sekg.ir.preprocessor.ngram import NGramPreprocessor
from sekg.util.annotation import exeTime

from sekg.graph.el.base import EntryPointLinker, EntityLinkResult
from sekg.constant.constant import OperationConstance, DomainConstant, WikiDataConstance, CodeConstant

import json
json.encoder.FLOAT_REPR = lambda x: format(x, '.3f')


class BestSeqSelector:

    def __init__(self, top_num=1):
        self.keyword2candidate_map = {}
        self.G = nx.Graph()
        self.best_selection = []
        self.best_score = []
        self.now_num = 0
        self.top_num = top_num

    def get_candidate_ids(self):
        result = set([])
        for keyword, ids in self.keyword2candidate_map.items():
            result = result | ids
        return list(result)

    def get_keyword_list(self):
        return list(self.keyword2candidate_map.keys())

    def add_candidate_for_keyword(self, keyword, candidate_node_id, score):
        if keyword not in self.keyword2candidate_map:
            self.keyword2candidate_map[keyword] = set([])

        self.keyword2candidate_map[keyword].add(candidate_node_id)
        self.add_candidate_entity_score(node_id=candidate_node_id, score=score)

    def add_candidate_entity_score(self, node_id, score):
        self.G.add_node(node_id, score=score)

    def add_pair_score(self, start_id, end_id, score):
        self.G.add_edge(start_id, end_id, score=score)

    def search_best_k_combination(self):
        self.chosen([], score=0.0, current_chosen_keyword_index=0)
        return self.get_best_k_selection()

    def search_best_combination(self):
        self.chosen([], score=0.0, current_chosen_keyword_index=0)
        return self.get_best_selection()

    def get_best_score(self):
        return self.best_score

    def get_best_selection(self):
        keyword_list = self.get_keyword_list()
        result = {}
        for keyword, node_id in zip(keyword_list, self.best_selection):
            result[keyword] = node_id[0]
        return result

    def get_best_k_selection(self):
        keyword_list = self.get_keyword_list()
        result = []
        for node_list in self.best_selection:
            temp = {}
            for keyword, node_id in zip(keyword_list, node_list):
                temp[keyword] = node_id
            result.append(temp)
        return result

    def insert(self, score, history_chosen, current_chosen_keyword_index):
        if self.now_num == 0:
            self.best_score.append(score)
            self.best_selection.append(history_chosen)
            self.now_num += 1
            return True
        else:
            tag = False
            score_index = 0
            while score_index < self.now_num:
                if self.best_score[score_index] < score:
                        self.best_score.insert(score_index, score)
                        self.best_selection.insert(score_index, history_chosen)
                        self.now_num += 1
                        tag = True
                        break
                score_index += 1
            if self.now_num >= self.top_num:
                self.now_num = self.top_num
                self.best_score = self.best_score[:self.top_num]
                self.best_selection = self.best_selection[:self.top_num]
            return tag

    def chosen(self, history_chosen, score, current_chosen_keyword_index):
        if current_chosen_keyword_index >= len(self.get_keyword_list()):
            tag = self.insert(score, history_chosen, current_chosen_keyword_index)
            if not tag:
                return
        else:
            current_keyword = self.get_keyword_list()[current_chosen_keyword_index]

            candidate_ids = self.keyword2candidate_map[current_keyword]

            for candidate_entity_id in candidate_ids:
                extra_score = self.G.nodes[candidate_entity_id]["score"]

                for history_chosen_entity_id in history_chosen:
                    extra_score += self.G[candidate_entity_id][history_chosen_entity_id]["score"]

                self.chosen(history_chosen + [candidate_entity_id], score=score + extra_score,
                            current_chosen_keyword_index=current_chosen_keyword_index + 1)


class PatternEntityPointLinker(EntryPointLinker):
    DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT = 3

    def __init__(self, preprocessor: NGramPreprocessor,
                 doc_sim_model: DocumentSimModel,
                 kg_name_searcher: KGNameSearcher,
                 graph2vecModel: Word2VecKeyedVectors,
                 graph_data: GraphData,
                 top_num=1
                 ):
        self.preprocessor = preprocessor
        self.doc_sim_model = doc_sim_model
        self.graph2vecModel = graph2vecModel
        self.kg_name_searcher = kg_name_searcher
        self.graph_data = graph_data
        self.lemmatizer = WordNetLemmatizer()
        # to start the util
        self.candidate_list = {}
        self.lemmatizer.lemmatize(word="test")
        self.selector = BestSeqSelector(top_num=top_num)

    def score_full_name(self, candidate_node_ids, keyword, keywords, name_score_map, original_query):
        new_keywords = set([])
        for keyword in keywords:
            new_keywords.add(keyword.lower())
        keywords = new_keywords
        clean_query_string = " ".join(list(keywords))
        # Step 1：计算备选集合之中出现的alias的TF-IDF值，将alias转化为n_array，用于便于后续相似度加成
        alias_list = []
        for node_id in candidate_node_ids:
            alias_list.append(list(self.kg_name_searcher.get_full_names(node_id)))
        dictionary = corpora.Dictionary(alias_list)
        new_corpus = [dictionary.doc2bow(text) for text in alias_list]
        tfidf = models.TfidfModel(new_corpus)

        # 计算 weight array
        weight_array = np.zeros(shape=(len(alias_list), len(dictionary)), dtype=float)
        for index, alias in enumerate(alias_list):
            string_bow = dictionary.doc2bow(alias)
            string_tfidf = tfidf[string_bow]
            for item in string_tfidf:
                weight_array[index, item[0]] = item[1]
        # Step 2：计算得到每个alias的分数向量, 取其加权平均数
        score_array = np.zeros(shape=(len(candidate_node_ids), len(dictionary)), dtype=float)
        for index, node_id in enumerate(candidate_node_ids):
            score = self.compute_position_name_score_for_one_node(
                keywords,
                dictionary,
                clean_query_string,
                node_id,
                original_query,
                len(dictionary)
            )
            if node_id not in name_score_map:
                if score > 0.5:
                    name_score_map[node_id] = score
            elif name_score_map[node_id] < score and score > 0.5:
                name_score_map[node_id] = score
        return name_score_map

    def compute_position_name_score_for_one_node(self, keywords, dictionary, query_string, node_id, original_query,
                                                 dim):
        full_names = self.kg_name_searcher.get_full_names(node_id)
        alia_vec = dictionary.doc2bow(full_names)
        score_vec = np.zeros(shape=(1, dim), dtype=float)
        for index, full_name in enumerate(full_names):
            full_name = self.lemmatizer.lemmatize(full_name.lower())
            query_string = self.lemmatizer.lemmatize(query_string.lower())
            original_query = self.lemmatizer.lemmatize(original_query.lower())
            name_set = set(full_name.split(" "))
            word_count = len(name_set)
            tokens = name_set & keywords
            token_score = 1.0 * len(tokens) / word_count

            appearing = 0.0
            for word in name_set:
                if word in query_string.split():
                    appearing += 1
                elif word in original_query.split():
                    appearing += 1

            appearing_score = appearing / word_count
            index_id = alia_vec[index][0]
            score_vec[0, index_id] = max([token_score, appearing_score])

        return max(list(score_vec[0, :]))

    def extract_keywords(self, query):
        keywords = self.preprocessor.get_n_gram_list(query)
        return keywords

    @exeTime
    def link(self, query, candidate_num=DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT):
        return self.link_n(query, top_num=1, candidate_num=candidate_num)[0]

    @exeTime
    def link_n(self, query, top_num=1, candidate_num=DEFAULT_CANDIDATE_NUM_FOR_EVERY_POINT):
        keywords = self.extract_keywords(query)
        selector = BestSeqSelector(top_num=top_num)
        doc_id_2_doc_retrieval_result_map = {}
        keyword_to_candidate_list_map = dict()

        for keyword in keywords:
            doc_retrieval_result_list = self.get_candidate_for_keyword(
                candidate_num=candidate_num,
                keyword=keyword,
                query=query,
                keywords=keywords,
                name_score_map={},
                node_type="n")
            for key, value in doc_retrieval_result_list.items():
                if len(value) > 0:
                    keyword_to_candidate_list_map[key] = value

        for keyword, doc_retrieval_result_list in keyword_to_candidate_list_map.items():

            for doc_retrieval_result in doc_retrieval_result_list:
                selector.add_candidate_for_keyword(keyword=keyword,
                                                   candidate_node_id=doc_retrieval_result.doc_id,
                                                   score=doc_retrieval_result.score)

                doc_id_2_doc_retrieval_result_map[doc_retrieval_result.doc_id] = doc_retrieval_result
        candidate_ids = selector.get_candidate_ids()

        vector_map = {}

        all_vectors = []

        new_ids = []

        for candidate_id in candidate_ids:
            try:
                vector_map[candidate_id] = self.graph2vecModel.wv[str(candidate_id)]
                all_vectors.append(vector_map[candidate_id])
                new_ids.append(candidate_id)
            except Exception:
                print("No node:", str(candidate_id))
                continue
        candidate_ids = new_ids

        if len(keywords) == 1:
            rate = 0.0
        else:
            rate = 2.0 / (len(list(set(keywords))) - 1.0)

        for start_id in candidate_ids:
            start_vector = vector_map[start_id]
            sim_vector = self.graph2vecModel.wv.cosine_similarities(start_vector, all_vectors)

            for end_id, score in zip(candidate_ids, sim_vector):
                selector.add_pair_score(start_id=start_id, end_id=end_id, score=score * rate)

        best_linking_result = selector.search_best_k_combination()
        result = []
        for index in range(top_num):
            result.append({})
        index = 0
        for node_list in best_linking_result:
            for keyword, node_id in node_list.items():
                doc_retrieval_result = doc_id_2_doc_retrieval_result_map[node_id]
                result[index][keyword] = EntityLinkResult(mention=keyword, node_id=node_id,
                                                          node_name=doc_retrieval_result.doc_name,
                                                          score=doc_retrieval_result.score)
            index += 1

        return result

    def get_candidate_for_keyword(self, candidate_num,
                                  keyword,
                                  query,
                                  name_score_map,
                                  keywords,
                                  node_type):

        new_keyword = self.lemmatizer.lemmatize(keyword.lower(), pos=node_type)
        word_list = new_keyword.split()
        full_name_match_id_set = self.kg_name_searcher.search_by_full_name(new_keyword)
        keyword_match_id_set = self.kg_name_searcher.search_by_keyword(new_keyword)
        if len(full_name_match_id_set | keyword_match_id_set) == 0 and len(word_list) > 1:
            link_result = {}
            for word in word_list:
                for key, value in self.get_candidate_for_keyword(candidate_num,
                                                                 word,
                                                                 query,
                                                                 name_score_map,
                                                                 keywords,
                                                                 node_type).items():
                    if key not in link_result:
                        link_result[key] = value
            return link_result
        else:
            candidate_doc_result_list = []
            # 全名搜索
            # get the operation id
            operation_ids = self.graph_data.get_node_ids_by_label(OperationConstance.LABEL_OPERATION)
            # get the domain id
            domain_term_ids = self.graph_data.get_node_ids_by_label(DomainConstant.LABEL_DOMAIN_TERM)
            wiki_terms_ids = self.graph_data.get_node_ids_by_label(WikiDataConstance.LABEL_WIKIDATA)
            code_terms_ids = self.graph_data.get_node_ids_by_label(CodeConstant.LABEL_CODE_ELEMENT)

            concept_term_ids = domain_term_ids | wiki_terms_ids

            if len(operation_ids & full_name_match_id_set) > 0:
                full_name_match_id_set = operation_ids & full_name_match_id_set
                keyword_match_id_set = operation_ids & keyword_match_id_set
            elif len(concept_term_ids & full_name_match_id_set) > 0:
                # search in domain:
                full_name_match_id_set = concept_term_ids & full_name_match_id_set
                keyword_match_id_set = concept_term_ids & keyword_match_id_set
            full_name_match_id_set = full_name_match_id_set - code_terms_ids
            keyword_match_id_set = keyword_match_id_set - full_name_match_id_set - code_terms_ids

            del operation_ids, domain_term_ids, wiki_terms_ids, concept_term_ids, code_terms_ids

            chosen_ids = set([])

            large_candidate_num = candidate_num + 20

            doc_retrieval_result_list = []

            if len(full_name_match_id_set) > 0:
                doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=large_candidate_num,
                                                                      valid_doc_id_set=full_name_match_id_set)

                for doc in doc_retrieval_result_list:
                    if doc.doc_id not in chosen_ids:
                        chosen_ids.add(doc.doc_id)
                        candidate_doc_result_list.append(doc)

            if len(keyword_match_id_set) > 0:
                doc_retrieval_result_list = self.doc_sim_model.search(query=query, top_num=large_candidate_num,
                                                                      valid_doc_id_set=keyword_match_id_set)

                for doc in doc_retrieval_result_list:
                    if doc.doc_id not in chosen_ids:
                        chosen_ids.add(doc.doc_id)
                        candidate_doc_result_list.append(doc)

            self.score_full_name(candidate_node_ids=list(chosen_ids), keyword=keyword,  keywords=keywords, name_score_map=name_score_map,
                                 original_query=query)

            # compute the total score of a node
            new_candicate_list = []
            for candidate_doc in candidate_doc_result_list:
                if candidate_doc.doc_id in name_score_map:
                    candidate_doc.extra_info["name_score"] = name_score_map[candidate_doc.doc_id]
                    candidate_doc.extra_info["doc_score"] = candidate_doc.score
                    candidate_doc.score = candidate_doc.extra_info["name_score"]
                    name = ""
                    if "term_name" in self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]:
                        name = self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]["term_name"]
                    elif "wikidata_name" in self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]:
                        name = self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]["wikidata_name"]
                    elif "name" in self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]:
                        name = self.graph_data.get_node_info_dict(candidate_doc.doc_id)["properties"]["name"]
                    candidate_doc.extra_info["token_score"] = fuzz.ratio(keyword, name) / 100
                    candidate_doc.score = candidate_doc.extra_info["name_score"] * 0.4 + \
                                          candidate_doc.extra_info["token_score"] * 0.6
                    new_candicate_list.append(candidate_doc)
                    new_candicate_list = sorted(new_candicate_list, key=lambda candidate_doc: candidate_doc.score,
                                                reverse=True)

            return {keyword: new_candicate_list[:candidate_num]}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
fetch wikadata and wikipedia item with api
"""

import asyncio
import pickle
import time
import traceback
from pathlib import Path
import os
import aiohttp
import async_timeout
from sekg.util.url_util import URLUtil
from sekg.util.annotation import catch_exception
from sekg.wiki.WikiDataItem import WikiDataItem


# if sys.platform == 'win32':
#     loop = asyncio.ProactorEventLoop()
#     asyncio.set_event_loop(loop)

# import wikipedia


class AsyncWikiSearcher:
    API_URL = 'https://www.wikidata.org/w/api.php'
    WIKI_PEDIA_URL = 'https://en.wikipedia.org/w/api.php'
    TYPE_ALL = 0
    TYPE_TITLE = 1
    TYPE_ITEM = 2
    TYPE_WIKIPEDIA = 3

    def __init__(self, proxy_server=None, pool_size=63, stride=60):
        self.proxy_server = proxy_server
        self.semaphore = asyncio.Semaphore(pool_size)
        self.title_cache = {}
        self.item_cache = {}
        self.wikipedia_content_cache = {}
        self.wikidata_wikipedia_cache = {}
        self.wikipedia_content_html_cache = {}

        self.stride = stride
        self.wiki_pedia_none_list = []
        self.wikidata_id_2_wikipedia_title_dic = {}
        self.wikipedia_title_2_wikidata_id_dic = {}

    def init_id_to_title(self):
        if self.item_cache:
            for id, item in self.item_cache.items():
                if id not in self.wikidata_id_2_wikipedia_title_dic.keys():
                    title = self.get_title_from_wikidata_item(item)
                    self.wikidata_id_2_wikipedia_title_dic[id] = title
                    if title not in self.wikipedia_title_2_wikidata_id_dic.keys():
                        self.wikipedia_title_2_wikidata_id_dic[title] = {id}
                    else:
                        self.wikipedia_title_2_wikidata_id_dic[title].add(id)

    async def __fetch_titles(self, query, limit=5):
        if query in self.title_cache:
            return self.title_cache[query]
        params = {
            'action': 'query',
            'format': 'json',
            'list': 'search',
            'srprop': 'snippet',
            'srlimit': limit,
            'srsearch': query
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.API_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            result = [
                                {"pageid": item["pageid"], "title": item["title"], "snippet": item.get("snippet", "")}
                                for item in json_data["query"]["search"]]
                            if result:
                                self.title_cache[query] = result
                            else:
                                print(query, ", no search result")
                            return result
        except Exception:
            print("[Failed] query: {}".format(query))
            traceback.print_exc()
            return []

    async def __fetch_entity(self, id):
        if id in self.item_cache:
            return self.item_cache[id]
        params = {
            'ids': id,
            'format': 'json',
            'action': 'wbgetentities'
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.API_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            wikidata_item = WikiDataItem(id, init_at_once=False).init_wikidata_item_from_json(
                                json_data)
                            self.item_cache[id] = wikidata_item
                            wikidata_title = self.get_title_from_wikidata_item(wikidata_item)
                            self.wikidata_id_2_wikipedia_title_dic[id] = wikidata_title
                            # print("[Done] title: {}".format(id))
                            return json_data
        except Exception:
            print("[Failed] title: {}".format(id))
            traceback.print_exc()
            return {}

    async def __fetch_wikipedia_content(self, title):
        if title in self.wiki_pedia_none_list:
            return {}
        if title in self.wikipedia_content_cache:
            return self.wikipedia_content_cache[title]
        params = {
            'titles': title,
            'format': 'json',
            'action': 'query',
            'prop': 'extracts',
            'exintro': '',
            'explaintext': '',
            'redirects': 1
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.WIKI_PEDIA_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            result = [
                                {"pageid": item["pageid"], "title": item["title"], "context": item.get("extract", ""),
                                 }
                                for index, item in json_data["query"]["pages"].items()]
                            if result:
                                self.wikipedia_content_cache[title] = result
                            else:
                                self.wiki_pedia_none_list.append(title)
                                print(title, ", no wikipedia context")
                            # print("[Done] title: {}".format(title))
                            return result
        except Exception:
            print("[Failed] title: {}".format(title))
            traceback.print_exc()
            return {}

    async def __fetch_wikipedia_context_html(self, title):
        if title in self.wiki_pedia_none_list:
            return {}
        if title in self.wikipedia_content_cache:
            return self.wikipedia_content_cache[title]
        params = {
            'titles': title,
            'format': 'json',
            'action': 'query',
            'prop': 'extracts',
            'exintro': '',
            'redirects': 1
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.WIKI_PEDIA_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            result = [
                                {"pageid": item["pageid"], "title": item["title"], "context": item.get("extract", ""),
                                 }
                                for index, item in json_data["query"]["pages"].items()]
                            if result:
                                self.wikipedia_content_html_cache[title] = result
                            else:
                                self.wiki_pedia_none_list.append(title)
                                print(title, ", no wikipedia context html")
                            # print("[Done] title: {}".format(title))
                            return result
        except Exception:
            print("[Failed] title: {}".format(title))
            traceback.print_exc()
            return {}

    async def __fetch_wikidata_id_by_wikipedia_title(self, title):
        if title in self.wikipedia_title_2_wikidata_id_dic.keys():
            return self.wikipedia_title_2_wikidata_id_dic[title]
        params = {
            'titles': title,
            'format': 'json',
            'action': 'query',
            'prop': 'pageprops',
            "ppprop": "wikibase_item",
            "redirects": 1
        }
        try:
            async with self.semaphore:
                async with aiohttp.ClientSession() as session:
                    with async_timeout.timeout(10):
                        async with session.get(self.WIKI_PEDIA_URL, params=params, proxy=self.proxy_server) as response:
                            json_data = await response.json()
                            result = set()
                            for key, item in json_data["query"]["pages"].items():
                                if "pageprops" in item.keys():
                                    wikidata_id = item["pageprops"].get("wikibase_item", "")
                                    result.add(wikidata_id)
                            if result:
                                self.wikipedia_title_2_wikidata_id_dic[title] = result
                            else:
                                print(title, ", no corresponding wikidata")

                            # print("[Done] title: {}".format(title))
                            return result
        except Exception:
            print("[Failed] title: {}".format(title))
            traceback.print_exc()
            return {}

    def __fet_wikipedia_content_by_wikidata_id(self, id):
        if id in self.wikidata_wikipedia_cache.keys():
            return self.wikidata_wikipedia_cache[id]
        wikipedia_title = self.get_wikipedia_title_by_wikidata_id(id)
        if wikipedia_title and wikipedia_title in self.wikipedia_content_cache.keys():
            wikipedia_context = self.wikipedia_content_cache[wikipedia_title]
            self.wikidata_wikipedia_cache[id] = wikipedia_context
            return wikipedia_context
        return {}

    def save(self, title_save_path=None, item_save_path=None, wikipedia_content_save_path=None,
             wikipedia_html_save_path=None, wikipedia_title_2_wikidata_id_path=None):
        print("Save...")
        if title_save_path is not None:
            with Path(title_save_path).open("wb") as f:
                pickle.dump(self.title_cache, f)
        if item_save_path is not None:
            with Path(item_save_path).open("wb") as f:
                pickle.dump(self.item_cache, f)
        if wikipedia_content_save_path is not None:
            with Path(wikipedia_content_save_path).open("wb") as f:
                pickle.dump(self.wikipedia_content_cache, f)
        if wikipedia_html_save_path is not None:
            with Path(wikipedia_html_save_path).open("wb") as f:
                pickle.dump(self.wikipedia_content_cache, f)
        if wikipedia_title_2_wikidata_id_path is not None:
            with Path(wikipedia_title_2_wikidata_id_path).open("wb") as f:
                pickle.dump(self.wikipedia_title_2_wikidata_id_dic, f)

    def save_by_type(self, save_path, type=0):
        if save_path:
            print("save....")
            if type == self.TYPE_ALL:
                save_data = {"title_cache": self.title_cache,
                             "item_cache": self.item_cache,
                             "wikipedia_context": self.wikipedia_content_cache
                             }
                with Path(save_path).open("wb") as f:
                    pickle.dump(save_data, f)
            elif type == self.TYPE_TITLE:
                with Path(save_path).open("wb") as f:
                    pickle.dump(self.title_cache, f)
            elif type == self.TYPE_ITEM:
                with Path(save_path).open("wb") as f:
                    pickle.dump(self.item_cache, f)
            elif type == self.TYPE_WIKIPEDIA:
                with Path(save_path).open("wb") as f:
                    pickle.dump(self.wikipedia_content_cache, f)
            else:
                pass

    def clear(self):
        del self.wikipedia_content_cache
        del self.item_cache
        del self.title_cache
        del self.wikidata_wikipedia_cache
        del self.wikipedia_content_html_cache

        self.item_cache = {}
        self.title_cache = {}
        self.wikipedia_content_cache = {}
        self.wikidata_wikipedia_cache = {}
        self.wikipedia_content_html_cache = {}

    def search_title(self, queries):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_titles(q) for q in queries]
        loop.run_until_complete(asyncio.gather(*tasks))
        return self.title_cache

    def fetch_item(self, ids):
        loop = asyncio.get_event_loop()
        if not ids:
            return {}
        else:
            tasks = [self.__fetch_entity(_id) for _id in ids]
            loop.run_until_complete(asyncio.gather(*tasks))
        return self.item_cache

    def fetch_item_neighbor(self, ids=None):
        neighbor_ids = set()
        self.fetch_item(ids)
        if ids is None:
            for _id, item in self.item_cache.items():
                neighbor_ids.update(item.get_neighbor_ids())
        else:
            for id in ids:
                if id in self.item_cache.keys():
                    item = self.item_cache[id]
                    neighbor_ids.update(item.get_neighbor_ids())
        self.fetch_item(neighbor_ids)
        return self.item_cache

    def fetch_wikipedia_context_for_wikidata(self, wikidata_ids):
        self.fetch_item(wikidata_ids)
        titles = [_title for _title in self.wikidata_id_2_wikipedia_title_dic.values() if _title]
        self.fetch_wikipedia_context(titles)
        for _id in wikidata_ids:
            self.__fet_wikipedia_content_by_wikidata_id(_id)
        return self.wikidata_wikipedia_cache

    def fetch_wikipedia_context(self, titles):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_wikipedia_content(_title) for _title in titles]
        loop.run_until_complete(asyncio.gather(*tasks))
        return self.item_cache

    def fetch_wikipedia_context_html(self, titles):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_wikipedia_context_html(_title) for _title in titles]
        loop.run_until_complete(asyncio.gather(*tasks))
        return self.item_cache

    def fetch_wikidata_id_by_wikipedia_title(self, titles):
        loop = asyncio.get_event_loop()
        tasks = [self.__fetch_wikidata_id_by_wikipedia_title(_title) for _title in titles]
        loop.run_until_complete(asyncio.gather(*tasks))
        return self.item_cache

    def fetch_wikidata_item_by_wikipedia_title(self, titles):
        self.fetch_wikidata_id_by_wikipedia_title(titles)
        print(self.wikipedia_title_2_wikidata_id_dic)
        wikidata_ids = set()
        for value in self.wikipedia_title_2_wikidata_id_dic.values():
            wikidata_ids.update(set(value))
        self.fetch_item(wikidata_ids)

    def get_title_from_wikidata_item(self, wikidata_item: WikiDataItem):
        url = wikidata_item.data_dict.get("site:enwiki", "")
        if url:
            wikidata_title = URLUtil.parse_url_to_title(url)
            return wikidata_title
        return ""

    def get_wikipedia_title_by_wikidata_id(self, id):
        if id in self.wikidata_id_2_wikipedia_title_dic:
            return self.wikidata_id_2_wikipedia_title_dic[id]
        return ""

    def update(self, cache_path, type=0):
        if cache_path is not None:
            if type == self.TYPE_ALL:
                self.init_from_cache_by_type(cache_path, type=type)
                self.save_by_type(cache_path, self.TYPE_ALL)
            elif type == self.TYPE_TITLE:
                self.init_from_cache(title_save_path=cache_path)
                self.save_by_type(cache_path, self.TYPE_TITLE)
            elif type == self.TYPE_ITEM:
                self.init_from_cache(item_save_path=cache_path)
                self.save_by_type(cache_path, self.TYPE_ITEM)
            elif type == self.TYPE_WIKIPEDIA:
                self.init_from_cache(wikipedia_save_path=cache_path)
                self.save_by_type(cache_path, self.TYPE_WIKIPEDIA)
            else:
                pass

    @catch_exception
    def init_from_cache_by_type(self, cache_path, type=0):
        if cache_path:
            print("Init from cache...")
            if type == self.TYPE_ALL:
                with Path(cache_path).open("rb") as f:
                    data = pickle.load(f)
                    self.title_cache = dict(self.title_cache, **data["title_cache"])
                    self.item_cache = dict(self.item_cache, **data["item_cache"])
                    self.wikipedia_content_cache = dict(self.wikipedia_content_cache, **data["wikipedia_context"])
                    self.init_id_to_title()
            elif type == self.TYPE_TITLE:
                self.init_from_cache(title_save_path=cache_path)
            elif type == self.TYPE_ITEM:
                self.init_from_cache(item_save_path=cache_path)
                self.init_id_to_title()
            elif type == self.TYPE_WIKIPEDIA:
                self.init_from_cache(wikipedia_save_path=cache_path)
            else:
                pass

    @catch_exception
    def init_from_cache(self, title_save_path=None, item_save_path=None, wikipedia_save_path=None,
                        wikidata_wikidedia_save_path=None, wikipedia_html_save_path=None,
                        wikipedia_title_2_wikidata_id_path=None):
        print("Init from cache...")
        if title_save_path is not None and Path(title_save_path).exists():
            with Path(title_save_path).open("rb") as f:
                title_cache = pickle.load(f)
                self.title_cache = dict(self.title_cache, **title_cache)
        if item_save_path is not None and Path(item_save_path).exists():
            with Path(item_save_path).open("rb") as f:
                item_cache = pickle.load(f)
                self.item_cache = dict(self.item_cache, **item_cache)
                self.init_id_to_title()
        if wikipedia_save_path is not None and Path(wikipedia_save_path).exists():
            with Path(wikipedia_save_path).open("rb") as f:
                wikipedia_cache = pickle.load(f)
                self.wikipedia_content_cache = dict(self.wikipedia_content_cache, **wikipedia_cache)
        if wikidata_wikidedia_save_path is not None and Path(wikidata_wikidedia_save_path).exists():
            with Path(wikidata_wikidedia_save_path).open("rb") as f:
                wikidata_wikipedia_cache = pickle.load(f)
                self.wikidata_wikipedia_cache = dict(self.wikidata_wikipedia_cache, **wikidata_wikipedia_cache)
        if wikipedia_html_save_path is not None and Path(wikipedia_html_save_path).exists():
            with Path(wikipedia_html_save_path).open("rb") as f:
                wikipedia_html_cache = pickle.load(f)
                self.wikipedia_content_html_cache = dict(self.wikipedia_content_html_cache, **wikipedia_html_cache)
        if wikipedia_title_2_wikidata_id_path is not None and Path(wikipedia_title_2_wikidata_id_path).exists():
            with Path(wikipedia_title_2_wikidata_id_path).open("rb") as f:
                wikipedia_title_2_wikidata_id_dic = pickle.load(f)
                self.wikipedia_title_2_wikidata_id_dic = dict(self.wikipedia_title_2_wikidata_id_dic,
                                                              **wikipedia_title_2_wikidata_id_dic)

    def item_cache_size(self):
        return len(self.item_cache)

    def title_cache_size(self):
        return len(self.title_cache)

    def wikipedia_cache_size(self):
        return len(self.wikipedia_content_cache)

    def wikidata_wikipedia_context_cache_size(self):
        return len(self.wikidata_wikipedia_cache)

    def wikidedia_context_html_cache(self):
        return len(self.wikipedia_content_html_cache)

    def id_to_title_cache_size(self):
        return len(self.wikidata_id_2_wikipedia_title_dic)

    def get_item_cache(self):
        return self.item_cache

    def get_title_cache(self):
        return self.title_cache

    def get_wikipedia_cache(self):
        return self.wikipedia_content_cache

    def get_wikidata_wikipedia_context_cache(self):
        return self.wikidata_wikipedia_cache

    def get_wikipedia_content_html(self):
        return self.wikipedia_content_html_cache

    def get_id_to_title_cache(self):
        return self.wikidata_id_2_wikipedia_title_dic

    def __repr__(self):
        return "<AsyncWikiSearcher title=%r item=%r wikipedia=%r>" % (
            self.title_cache_size(), self.item_cache_size(), self.wikipedia_cache_size())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: isky
@Email: 19110240019@fudan.edu.cn
@Created: 2019/12/13
------------------------------------------
@Modify: 2019/12/13
------------------------------------------
@Description:
"""
import neuralcoref
import spacy

from sekg.text.spacy_pipeline.tokenizer import CustomizeSpacy
from sekg.text.spacy_pipeline.sentenceHandler import SentenceHandler
from sekg.util.spacy_fixer import SoftwareTextPOSFixer


class PipeLineFactory:
    @staticmethod
    def custom_pipeline(**config):
        """
        create a new custom pipeline with fixing some tokenization problem.
        """
        nlp = spacy.load('en_core_web_sm')
        CustomizeSpacy.customize_tokenizer_split_single_lowcase_letter_and_period(nlp)
        CustomizeSpacy.customize_tokenizer_merge_hyphen(nlp)
        CustomizeSpacy.customize_tokenizer_api_name_recognition(nlp)
        nlp.add_pipe(SentenceHandler.hyphen_handler, name='sh', after='tagger')
        return nlp

    @staticmethod
    def full_pipeline():
        nlp = spacy.load('en', disable=["ner"])
        CustomizeSpacy.customize_tokenizer_split_single_lowcase_letter_and_period(nlp)
        CustomizeSpacy.customize_tokenizer_merge_hyphen(nlp)
        CustomizeSpacy.customize_tokenizer_api_name_recognition(nlp)
        nlp.add_pipe(SentenceHandler.hyphen_handler, name='sh', after='tagger')
        # 将核心动词与紧挨着的head是这个动词的介词合并
        # nlp.add_pipe(SentenceHandler.merge_verb_prep, name='merge_verb_prep', after='parser')
        nlp.add_pipe(SoftwareTextPOSFixer.fixer_for_pos, name="pos_fixer", after="tagger")
        nlp.add_pipe(SentenceHandler.tag_fixer_after_parser, name='pos_fixer_after_parser', after='parser')
        return nlp

    @staticmethod
    def full_pipeline_for_coreference_resolution():
        nlp = spacy.load('en', disable=["ner"])

        nlp.add_pipe(SentenceHandler.hyphen_handler, name='sh', before='tagger')
        nlp.add_pipe(SoftwareTextPOSFixer.fixer_for_pos, name="pos_fixer", after="tagger")
        neuralcoref.add_to_pipe(nlp, greedyness=0.5)
        return nlp



#!/usr/bin/python
# -*- coding: UTF-8 -*-
import os

import gensim

from sekg.ir.doc.wrapper import PreprocessMultiFieldDocumentCollection
from sekg.ir.models.base import DocumentSimModel
from sekg.util.file import DirUtil
import numpy as np


class LDAModel(DocumentSimModel):
    __lda_model_name__ = "lda.model"
    __corpus_name__ = "corpus.mm"
    __dictionary_name__ = "dictionary.dict"
    __lda_sim_index__ = "LDA.sim.index"

    __entity_collection_name__ = "doc.pre.collection"
    DEFAULT_TOPIC_NUM = 1000

    def __init__(self, name, model_dir_path, **config):

        """
        init the lsi model with
        :param model_dir_path:
        """
        super().__init__(name, model_dir_path, **config)
        self.lda_model = None
        self.corpus = None
        self.dict = None
        self.similarity_lda_index = None
        self.__init_sub_model_path__()
        self.num_topics = LDAModel.DEFAULT_TOPIC_NUM
        self.preprocessor = None

        self.preprocess_doc_collection = None

    def __init_sub_model_path__(self):
        if self.model_dir_path is None:
            return
        # init the paths of models
        self.lda_model_path = os.path.join(self.model_dir_path, self.__lda_model_name__)
        self.corpus_path = os.path.join(self.model_dir_path, self.__corpus_name__)
        self.dictionary_path = os.path.join(self.model_dir_path, self.__dictionary_name__)

        self.entity_collection_path = os.path.join(self.model_dir_path, self.__entity_collection_name__)

        self.sim_index_dir = os.path.join(self.model_dir_path, "index")

        self.sim_index_path = os.path.join(self.sim_index_dir, self.__lda_sim_index__)

        DirUtil.create_file_dir(self.model_dir_path)
        DirUtil.create_file_dir(self.sim_index_dir)

    def init_model(self):
        """
               init the model
               :return:
               """
        self.__init_sub_model_path__()
        print("loading the LDA models")
        self.lda_model = gensim.models.LdaModel.load(self.lda_model_path)
        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load LSI models done")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.__init_document_collection(preprocess_doc_collection)
        self.__init_index()

    def init_model_as_submodel(self):
        """
        init the model
        :return:
        """
        self.__init_sub_model_path__()
        print("loading the LDA models")
        self.lda_model = gensim.models.LdaModel.load(self.lda_model_path)
        self.corpus = gensim.corpora.MmCorpus(self.corpus_path)
        self.dict = gensim.corpora.Dictionary.load(self.dictionary_path)
        print("All document number: ", self.dict.num_docs)
        print("All words number: ", self.dict.num_pos)
        print("load LSI models done")
        preprocess_doc_collection = PreprocessMultiFieldDocumentCollection.load(self.entity_collection_path)
        self.preprocessor = preprocess_doc_collection.get_preprocessor()
        self.field_set = self.preprocess_doc_collection.get_field_set()
        self.__init_index()

    def __init_document_collection(self, preprocess_doc_collection):
        self.preprocess_doc_collection = preprocess_doc_collection
        self.preprocessor = self.preprocess_doc_collection.get_preprocessor()
        self.field_set = self.preprocess_doc_collection.get_field_set()

    def train_from_document_collection(self, preprocess_document_collection, num_topics=400, **config):
        print("start training")
        self.num_topics = num_topics
        self.__init_document_collection(preprocess_document_collection)
        corpus_clean_text = []
        preprocess_multi_field_doc_list = preprocess_document_collection.get_all_preprocess_document_list()

        for docno, multi_field_doc in enumerate(preprocess_multi_field_doc_list):
            corpus_clean_text.append(multi_field_doc.get_document_text_words())
            # print("process doc %d" % docno)

        print("corpus len=%d" % len(corpus_clean_text))

        print("Dictionary init...")

        self.dict = gensim.corpora.Dictionary(corpus_clean_text)
        print("Dictionary init complete ")

        self.corpus = [self.dict.doc2bow(text) for text in corpus_clean_text]

        print("LDA Training...")
        alpha = 'symmetric'
        eta = None
        if "alpha" in config.keys():
            alpha = config["alpha"]
        if "eta" in config.keys():
            eta = config["eta"]

        self.lda_model = gensim.models.LdaModel(corpus=self.corpus,
                                                id2word=self.dict,
                                                num_topics=num_topics,
                                                alpha=alpha,
                                                eta=eta)
        print("LDA Train complete")

    def train_from_doc_collection_with_preprocessor(self, doc_collection, num_topics=400, **config):
        # TODO: FIX THE PARAMETER, MAKE THE PARAMETER INPUT COULD BE MULTIPLE CHOICE
        if isinstance(doc_collection, PreprocessMultiFieldDocumentCollection):
            alpha = 'symmetric'
            eta = None
            if "alpha" in config.keys():
                alpha = config["alpha"]
            if "eta" in config.keys():
                eta = config["eta"]
            self.train_from_document_collection(preprocess_document_collection=doc_collection, num_topics=num_topics,
                                                alpha=alpha, eta=eta)

    def save(self, model_dir_path):
        """
        save model to the model_dir_path
        :param model_dir_path: the dir to save the model
        :return:
        """
        super().save(model_dir_path)
        self.__init_sub_model_path__()

        self.dict.save(self.dictionary_path)

        print("build dictionary in %s" % self.dictionary_path)

        gensim.corpora.MmCorpus.serialize(self.corpus_path, self.corpus)
        print("save the corpus to %s" % self.corpus_path)

        self.lda_model.save(self.lda_model_path)
        print("LDA Training finish , save to %s" % self.lda_model_path)

        print("LDA similarity index init ...")
        self.similarity_lda_index = gensim.similarities.Similarity(self.sim_index_path,
                                                                   corpus=self.lda_model[self.corpus],
                                                                   num_features=self.num_topics)
        print("LDA similarity index complete")

        self.similarity_lda_index.save(self.sim_index_path)
        print("LDA similarity index finish , save to %s" % self.sim_index_path)

        print("entity collection saving...")
        self.preprocess_doc_collection.save(self.entity_collection_path)
        print(
            "entity collection finish saving , save to %s, %r" % (
                self.entity_collection_path, self.preprocess_doc_collection))

    def get_full_doc_score_vec(self, query):
        full_entity_score_vec = self.get_cache_score_vector(query)
        if full_entity_score_vec is not None:
            return full_entity_score_vec

        query_words = self.preprocessor.clean(query)
        query_bow, query_oov_dict = self.dict.doc2bow(query_words, return_missing=True)
        query_lda_vec = self.lda_model[query_bow]
        full_entity_score_vec = self.similarity_lda_index[query_lda_vec]
        self.cache_entity_score_vector(query, full_entity_score_vec)

        return full_entity_score_vec

    def __init_index(self):
        try:
            self.similarity_lda_index = gensim.similarities.Similarity.load(self.sim_index_path)
            self.get_full_doc_score_vec("test")

        except FileNotFoundError:
            print("the sim index is useless, generate again")
            self.similarity_lda_index = gensim.similarities.Similarity(self.sim_index_path,
                                                                       corpus=self.lda_model[self.corpus],
                                                                       num_features=self.num_topics)
            self.similarity_lda_index.save(self.sim_index_path)
